##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase

ZopeTestCase.installProduct('ParsedXML')

# FIXME: could test with Core DOM instead, do we want to?
from Products.ParsedXML import ParsedXML, PrettyPrinter
from StringIO import StringIO

def printElement(element):
    output = StringIO()
    PrettyPrinter.PrintVisitor(element, output)()
    return output.getvalue()

def checkOutput(wanted, got, message="Bad output"):
    assert wanted == got, \
           ("%s.  Wanted:\n%s[[EOF]]\nGot:\n%s[[EOF]]\n" 
            % (message, wanted, got))


class PrintTestBase(ZopeTestCase.ZopeTestCase):
    implementation = ParsedXML.theDOMImplementation

    def parse(self, xml):
        return ParsedXML.ParsedXML('foo', xml)


class PrintTestCase(PrintTestBase):

    def testAttrOrder(self):
        inStr = '<?xml version="1.0" ?>\n<doc a1="v1" a2="v2" a3="v3"/>\n'
        doc = self.parse(inStr)
        output = printElement(doc)

        checkOutput(inStr, output, "attribute order not preserved")

    def testDefaultAttrSkipped(self):
        inStr = '<!DOCTYPE doc [<!ELEMENT doc EMPTY>' \
                '<!ATTLIST doc a1 CDATA "v1">]><doc></doc>'
        outStr = '<?xml version="1.0" ?>\n<doc/>\n'
        doc = self.parse(inStr)
        output = printElement(doc)

        checkOutput(outStr, output, "default attribute printed")

    def testAttrEntRefExpansion(self):
        "entity references must be expanded in attributes"
        doc = self.implementation.createDocument(None, 'root', None)
        attr = doc.createAttribute("attrName")
        # this is a text string; &test; is not an entity reference,
        # so its & should be converted to the proper ref on printing,
        # as should the & and < and >.
        attr.value = "&test; &<>"
        # this reference should be expanded to the empty string
        attr.appendChild(doc.createEntityReference('foo'))
        doc.documentElement.setAttributeNode(attr)

        outStr = '<root attrName="&amp;test; &amp;&lt;&gt;"/>'
        output = printElement(doc.documentElement)
        checkOutput(outStr, output, "improper attr entity expansion")

    def testTextEntRefExpansion(self):
        "only some entity references should be expanded in text"
        doc = self.implementation.createDocument(None, 'root', None)
        # &< must be converted to the proper reference; > should not.
        text = doc.createTextNode("&<>]]>")
        outStr = '&amp;&lt;>]]&gt;'
        output = printElement(text)
        checkOutput(outStr, output, "improper text entity expansion")

    #TODO: check for expansion of entity refs in other contexts;
    #currently we're expanding aggressively, but it's not a priority
    #because the parser gets to play around with refs too

class Lvl2PrintTestCase(PrintTestBase):

    def testNamespacePrint(self):
        outStr = '<?xml version="1.0" ?>\n' \
                    '<bar xmlns:foo="uri:foo">\n' \
                    '<foo:baz/></bar>\n'
        doc = self.parse(outStr)
        output = printElement(doc)
        checkOutput(outStr, output)

    def testNamespaceAttrOrder(self):
        inStr = ('<?xml version="1.0" ?>\n'
                  '<doc>'
                  '  <fooE xmlns="defaultURL" xmlns:oneN="oneURL" '
                          'xmlns:twoN="twoURL" xmlns:threeN="threeURL"/>'
                  '</doc>\n')
        doc = self.parse(inStr)
        output = printElement(doc)
        checkOutput(inStr, output, "attribute order not preserved")


    def testHierarchicalElementNamespacePrint(self):
        # print new ns, don't print ns printed by ancestor
        outStr = ('<?xml version="1.0" ?>\n'
                  '<fooN:fooE xmlns:fooN="fooURL">'
                  '  <barN:barE xmlns:barN="barURL">'
                  '  <fooN:fooE>'
                  '    <fooN:fooE/>'
                  '  </fooN:fooE>'
                  '  </barN:barE>'
                  '</fooN:fooE>\n')
        doc = self.parse(outStr)
        output = printElement(doc)
        checkOutput(outStr, output)

    def testDefaultNamespacePrint(self):
        outStr = ('<?xml version="1.0" ?>\n'
                  '<fooE xmlns="defaultURL"><fooE xmlns="barURL"/>'
                  '</fooE>\n')
        doc = self.parse(outStr)
        output = printElement(doc)
        checkOutput(outStr, output)

    def testDefaultAndPrefixNamespacePrint(self):
        # try and tickle a namespace printing bug
        outStr = ('<?xml version="1.0" ?>\n'
                  '<doc>'
                  '  <fooE xmlns="defaultURL" xmlns:fooN="fooURL" '
                          'xmlns:barN="barUrl" xmlns:bazN="bazUrl">'
                  '  <barE bazN:bazE="baz"/>'
                  '  <bazN:bazE/>'
                  '  <quxN xmlns:quxE="quxUrl"/>'
                  '  <quxN xmlns:quxE="quxUrl" bazN:bazE="baz"/>'
                  '  </fooE>'
                  '</doc>\n')
        doc = self.parse(outStr)
        output = printElement(doc)
        checkOutput(outStr, output)

if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(PrintTestCase))
        suite.addTest(unittest.makeSuite(Lvl2PrintTestCase))
        return suite



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Implementation of DOM Level 2 Traversal.

Based on the W3C recommendation at:
    http://www.w3.org/TR/DOM-Level-2-Traversal-Range/
"""

# This code could be sped up.
# - uses DOM methods, could use implementation internals, esp. childNodes
# - if we had mutation events, NodeIterator could build an array as it
#   iterated, and move over that, only updating on mutation

import xml.dom


__all__ = [
    "NodeFilter",
    "NodeIterator",
    "TreeWalker",
    ]


class NodeFilter:
    # Constants returned by acceptNode():
    FILTER_ACCEPT                  = 1
    FILTER_REJECT                  = 2
    FILTER_SKIP                    = 3

    # Constants for whatToShow:
    SHOW_ALL                       = 0xFFFFFFFF
    SHOW_ELEMENT                   = 0x00000001
    SHOW_ATTRIBUTE                 = 0x00000002
    SHOW_TEXT                      = 0x00000004
    SHOW_CDATA_SECTION             = 0x00000008
    SHOW_ENTITY_REFERENCE          = 0x00000010
    SHOW_ENTITY                    = 0x00000020
    SHOW_PROCESSING_INSTRUCTION    = 0x00000040
    SHOW_COMMENT                   = 0x00000080
    SHOW_DOCUMENT                  = 0x00000100
    SHOW_DOCUMENT_TYPE             = 0x00000200
    SHOW_DOCUMENT_FRAGMENT         = 0x00000400
    SHOW_NOTATION                  = 0x00000800

    def acceptNode(self, node):
        # Just accept everything by default:
        return NodeFilter.FILTER_ACCEPT


_whatToShow_bits = (
    (xml.dom.Node.ELEMENT_NODE, NodeFilter.SHOW_ELEMENT),
    (xml.dom.Node.ATTRIBUTE_NODE, NodeFilter.SHOW_ATTRIBUTE),
    (xml.dom.Node.TEXT_NODE, NodeFilter.SHOW_TEXT),
    (xml.dom.Node.CDATA_SECTION_NODE, NodeFilter.SHOW_CDATA_SECTION),
    (xml.dom.Node.ENTITY_REFERENCE_NODE, NodeFilter.SHOW_ENTITY_REFERENCE),
    (xml.dom.Node.ENTITY_NODE, NodeFilter.SHOW_ENTITY),
    (xml.dom.Node.PROCESSING_INSTRUCTION_NODE,
     NodeFilter.SHOW_PROCESSING_INSTRUCTION),
    (xml.dom.Node.COMMENT_NODE, NodeFilter.SHOW_COMMENT),
    (xml.dom.Node.DOCUMENT_NODE, NodeFilter.SHOW_DOCUMENT),
    (xml.dom.Node.DOCUMENT_TYPE_NODE, NodeFilter.SHOW_DOCUMENT_TYPE),
    (xml.dom.Node.DOCUMENT_FRAGMENT_NODE, NodeFilter.SHOW_DOCUMENT_FRAGMENT),
    (xml.dom.Node.NOTATION_NODE, NodeFilter.SHOW_NOTATION),
    )

class AccessorBase:
    def __init__(self, root, whatToShow, filter, entityReferenceExpansion):
        if root is None:
            raise xml.dom.NotSupportedErr(
                "root of traversal object can't be None")
        d = self.__dict__
        d['root'] = root
        d['whatToShow'] = whatToShow
        d['filter'] = filter
        d['expandEntityReferences'] = entityReferenceExpansion
        #
        # Decode the whatToShow flags for faster tests; the W3C
        # reserves the first 200 NodeType values, but the whatToShow
        # flags have to fit in 32 bits (leave slot 0 empty since it's
        # not a valid NodeType).
        #
        d['_whatToShow'] = what = [0] * 33
        for nodeType, bit in _whatToShow_bits:
            what[nodeType] = whatToShow & bit

    def __setattr__(self, name, value):
        setter = getattr(self, '_set_' + name, None)
        if setter is None:
            getter = getattr(self, '_get_' + name, None)
            if getter:
                raise xml.dom.NoModificationAllowedErr(
                    "read-only attribute: " + `name`)
            else:
                raise AttributeError, "no such attribute: " + `name`
        setter(value)

    def _get_root(self):
        return self.root

    def _get_whatToShow(self):
        return self.whatToShow

    def _get_filter(self):
        return filter

    def _get_expandEntityReferences(self):
        return self.expandEntityReferences

    def _should_show(self, node):
        if not self._whatToShow[node.nodeType]:
            return NodeFilter.FILTER_SKIP
        else:
            if (  node.nodeType == xml.dom.Node.ENTITY_REFERENCE_NODE
                  and not self.expandEntityReferences):
                return NodeFilter.FILTER_REJECT
            elif self.filter is not None:
                return self._filterNode(node)
        return NodeFilter.FILTER_ACCEPT

    def _nextInTree(self, node):
        """Return first visible node in node's subtree, or None."""
        # check given node first
        if self._should_show(node) == NodeFilter.FILTER_ACCEPT:
            return node
        elif self._should_show(node) == NodeFilter.FILTER_REJECT:
            return None
        for c in node.childNodes:
            child = self._nextInTree(c)
            if child:
                return child
            if c.isSameNode(self.root): # don't leave root subtree
                return None
        return None

    def _lastInTree(self, node):
        """Return last visible node in node's subtree, or None."""
        if self._should_show(node) == NodeFilter.FILTER_REJECT:
            return None
        childNodes = node.childNodes
        childNums = range(childNodes.length)
        childNums.reverse()
        for c in childNums:
            childNode = childNodes[c]
            child = self._lastInTree(childNode)
            if child:
                return child
            if childNode.isSameNode(self.root): # don't leave root subtree
                return None
        # subtree exhausted, check given node
        if self._should_show(node) == NodeFilter.FILTER_ACCEPT:
            return node
        return None

    # we don't do any visibilty tests here, _nextInTree does.
    def _nextNode(self, startNode):
        """Return next visible node after startNode, or None."""
        # check children
        for child in startNode.childNodes:
            node = self._nextInTree(child)
            if node:
                return node
            if child.isSameNode(self.root): # don't leave root subtree
                return None
        # check next siblings
        sib = startNode.nextSibling
        while sib:
            node = self._nextInTree(sib)
            if node:
                return node
            if sib.isSameNode(self.root): # don't leave root subtree
                return None
            sib = sib.nextSibling
        # check ancestors' next siblings; don't visit ancestors
        ancestor = startNode.parentNode
        while ancestor:
            sib = ancestor.nextSibling
            while sib:
                node = self._nextInTree(sib)
                if node:
                    return node
                if sib.isSameNode(self.root): # don't leave root subtree
                    return None
                sib = sib.nextSibling
            # no visible nodes in siblings or subtrees of this ancestor
            if ancestor.isSameNode(self.root):
                # don't leave root subtree                
                return None
            ancestor = ancestor.parentNode
        return None

    # we *do* a visibilty test here, _lastInTree does too.
    def _previousNode(self, startNode):
        """Return the previous visible node after startNode, or None."""
        # check previous siblings
        sib = startNode.previousSibling
        while sib:
            node = self._lastInTree(sib)
            if node:
                return node
            if sib.isSameNode(self.root): # don't leave root subtree
                return None
            sib = sib.previousSibling
        # check ancestors, then ancestors' previous siblings
        ancestor = startNode.parentNode
        while ancestor:
            if self._should_show(ancestor) == NodeFilter.FILTER_ACCEPT:
                return ancestor
            sib = ancestor.previousSibling
            while sib:
                node = self._lastInTree(sib)
                if node:
                    return node
                if sib.isSameNode(self.root): # don't leave root subtree
                    return None
                sib = sib.previousSibling
            if ancestor.isSameNode(self.root):
                # don't leave root subtree
                return None
            ancestor = ancestor.parentNode
        return None

# Since we don't need to know about structure, we could probably be a lot
# faster if we kept a list of nodes in document order and updated
# it when we got a mutation event - once we have mutation events.
class NodeIterator(AccessorBase):
    BEFORE_NODE = 1 # iterator crossed reference node moving forward
    AFTER_NODE = 0  # iterator crossed reference node moving backward

    def __init__(self, root, whatToShow=NodeFilter.SHOW_ALL, filter=None,
                 entityReferenceExpansion=1):
        AccessorBase.__init__(self, root, whatToShow,
                              filter, entityReferenceExpansion)
        self.__dict__['_refNode'] = None
        self.__dict__['_refPos'] = NodeIterator.BEFORE_NODE

    def detach(self):
        self.__dict__['root'] = None

    def nextNode(self):
        if self.root is None:
            raise xml.dom.InvalidStateErr(
                "can't iterate using a detached NodeIterator")
        if self._refNode == None:
            self.__dict__['_refNode'] = self.root
            self.__dict__['_refPos'] = NodeIterator.AFTER_NODE
            if self._should_show(self._refNode) == NodeFilter.FILTER_ACCEPT:
                return self._refNode
        elif self._refPos == NodeIterator.BEFORE_NODE:
            if self._should_show(self._refNode) == NodeFilter.FILTER_ACCEPT:
                self.__dict__['_refPos'] = NodeIterator.AFTER_NODE
                return self._refNode
        
        node = AccessorBase._nextNode(self, self._refNode)
        if node:
            self.__dict__['_refNode'] = node
        self.__dict__['_refPos'] = NodeIterator.AFTER_NODE
        return node

    def previousNode(self):
        if self.root is None:
            raise xml.dom.InvalidStateErr(
                "can't iterate using a detached NodeIterator")
        if self._refNode == None:
            self.__dict__['_refNode'] = self.root
            self.__dict__['_refPos'] = NodeIterator.BEFORE_NODE
        elif self._refPos == NodeIterator.AFTER_NODE:
            if self._should_show(self._refNode) == NodeFilter.FILTER_ACCEPT:
                self.__dict__['_refPos'] = NodeIterator.BEFORE_NODE
                return self._refNode
        
        node = AccessorBase._previousNode(self, self._refNode)
        if node:
            self.__dict__['_refNode'] = node            
        self.__dict__['_refPos'] = NodeIterator.BEFORE_NODE        
        return node

    def __getitem__(self, index):
        node = self.nextNode()
        if node is None:
            raise IndexError, "NodeIterator index out of range"
        return node

    def _filterNode(self, node):
        """Return what the filter says to do with node,
        translating reject into skip"""
        filterAction = self.filter.acceptNode(node)
        if filterAction == NodeFilter.FILTER_REJECT:
            return NodeFilter.FILTER_SKIP
        return filterAction


class TreeWalker(AccessorBase):
    def __init__(self, root, whatToShow=NodeFilter.SHOW_ALL, filter=None,
                 entityReferenceExpansion=1):
        AccessorBase.__init__(self, root, whatToShow,
                              filter, entityReferenceExpansion)
        self.__dict__['currentNode'] = root
        
    def _get_currentNode(self):
        return self.currentNode

    def _set_currentNode(self, node):
        if node is None:
            raise xml.dom.NotSupportedErr("can't set current node to None")
        self.__dict__['currentNode'] = node

    def parentNode(self):
        if self.root.isSameNode(self.currentNode):
            return None
        node = self.currentNode.parentNode
        while node is not None and (
            self._should_show(node) != NodeFilter.FILTER_ACCEPT):
            if node.isSameNode(self.root):
                # can't step any further up
                return
            else:
                node = node.parentNode
        if node is not None:
            self.__dict__['currentNode'] = node
        return node

    def firstChild(self):
        node = self.currentNode.firstChild
        while node is not None and (
            self._should_show(node) != NodeFilter.FILTER_ACCEPT):
            node = node.nextSibling
        if node is not None:
            self.__dict__['currentNode'] = node
        return node

    def lastChild(self):
        node = self.currentNode.lastChild
        while node is not None and (
            self._should_show(node) != NodeFilter.FILTER_ACCEPT):
            node = node.previousSibling
        if node is not None:
            self.__dict__['currentNode'] = node
        return node

    # the rec doesn't say that *Sibling should pay attention to root!
    def previousSibling(self):
        node = self.currentNode.previousSibling
        while node is not None and (
            self._should_show(node) != NodeFilter.FILTER_ACCEPT):
            node = node.previousSibling
        if node is not None:
            self.__dict__['currentNode'] = node
        return node

    def nextSibling(self):
        node = self.currentNode.nextSibling
        while node is not None and (
            self._should_show(node) != NodeFilter.FILTER_ACCEPT):
            node = node.nextSibling
        if node is not None:
            self.__dict__['currentNode'] = node
        return node

    # TreeWalkers don't move if there is no visible next or previous,
    # so we do nothing for a None return.
    def nextNode(self):
        node = AccessorBase._nextNode(self, self.currentNode)
        if node:
            self.__dict__['currentNode'] = node
        return node

    def previousNode(self):
        node = AccessorBase._previousNode(self, self.currentNode)
        if node:
            self.__dict__['currentNode'] = node
        return node

    def _filterNode(self, node):
        """Return what the filter says to do with node."""
        return self.filter.acceptNode(node)



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""
A simple example implementation of DOMProxy.

Provide a class to implement wrapDOMObj(), mix that and the DOMProxy proxy
classes, and add whatever you want to make it interesting.
"""

import DOMProxy
import Acquisition

class TransPubWrapper:
    """
    Mixin class to go alongside DOMProxy classes.  Provides the wrapDOMObj
    function to create TransPubNode classes.  This is what makes TransPubNodes
    transient - we never re-use proxies, but create new ones when we need them.
    """
    
    # In the future we may want to return different proxy types based
    # on non-DOM node types, such as DB/nonDB.  Probably map
    # the same way that the DOM classes will map.
    def wrapDOMObj(self, node):
        """
        Return the appropriate manageable class wrapped around the Node.
        We never create Nodes ourselves, only wrap existing ones.
        Wrapped node can be single DOM object, a non-DOM object, or a
        container that contains only non-DOM objects - DOM objects in
        containters aren't wrapped.
        """
        from xml.dom import Node
        import types
        if node == None:
            return None
        elif isinstance(node, types.InstanceType) \
             and node.__class__.__name__ == "ChildNodeList": # XXX impl detail
            return TransPubNodeList(node)
        elif isinstance(node, types.InstanceType) \
             and node.__class__.__name__ == "AttributeMap": # XXX impl detail
            return TransPubNamedNodeMap(node)
        elif not hasattr(node, "nodeType"):
            return node                     # not DOM, don't wrap.
        elif node.nodeType == Node.ELEMENT_NODE:
            return TransPubElement(node)
        elif node.nodeType == Node.ATTRIBUTE_NODE:
            return TransPubAttr(node)
        elif node.nodeType == Node.TEXT_NODE:
            return TransPubText(node)
        elif node.nodeType == Node.CDATA_SECTION_NODE:
            return TransPubCDATASection(node)
        elif node.nodeType == Node.ENTITY_REFERENCE_NODE:
            return TransPubEntityReference(node)
        elif node.nodeType == Node.ENTITY_NODE:
            return TransPubEntity(node)
        elif node.nodeType == Node.PROCESSING_INSTRUCTION_NODE:
            return TransPubProcessingInstruction(node)
        elif node.nodeType == Node.COMMENT_NODE:
            return TransPubComment(node)
        elif node.nodeType == Node.DOCUMENT_NODE:
            return TransPubDocument(node)
        elif node.nodeType == Node.DOCUMENT_TYPE_NODE:
            return TransPubDocumentType(node)
        elif node.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
            return TransPubDocumentFragment(node)
        elif node.nodeType == Node.NOTATION_NODE:
            return TransPubNotation(node)
        else:
            raise TypeError


_TRANS_PUB_DOM_PROXY_FEATURES = (
    ("org.zope.dom.acquisition", None),
    ("org.zope.dom.acquisition", "1.0"),
    )

class TransPubDOMImplementation(DOMProxy.DOMImplementationProxy):
    """
    A DOMImplementation proxy that implements createDocument to produce
    TransPubDocument instances.
    """

    def hasFeature(self, feature, version):
        feature = string.lower(feature)
        if (feature, version) in _TRANS_PUB_DOM_PROXY_FEATURES:
            return 1
        return self._domimplementation.hasFeature(feature, version)
    
    def createDocument(self, namespaceURI, qualifiedName, docType=None):
        DOMDocument = self._createDOMDocument(namespaceURI,
                                              qualifiedName, docType)
        return TransPubDocument(DOMDocument.aq_base) # XXX check aq

theDOMImplementation = TransPubDOMImplementation()

#DOMIO, DOMManageable, DOMPublishable,
class TransPubNode(TransPubWrapper, DOMProxy.NodeProxy, Acquisition.Implicit):
    "The core of the TransPub DOM proxies."
    pass

class TransPubNodeList(TransPubWrapper, DOMProxy.NodeListProxy):    
    "A TransPubWrapper mixer with NodeListProxy."
    pass

class TransPubNamedNodeMap(TransPubWrapper, DOMProxy.NamedNodeMapProxy):
    "A TransPubWrapper mixer with NamedNodeMapProxy."
    pass
    
class TransPubDocumentFragment(DOMProxy.DocumentFragmentProxy, TransPubNode):
    "A TransPubWrapper mixer with DocumentFragmentProxy."
    pass
    
class TransPubElement(DOMProxy.ElementProxy, TransPubNode):
    "A TransPubWrapper mixer with ElementProxy."
    pass

class TransPubCharacterData(DOMProxy.CharacterDataProxy, TransPubNode):
    "A TransPubWrapper mixer with CharacterDataProxy."
    pass

class TransPubCDATASection(DOMProxy.CDATASectionProxy, TransPubNode):
    "A TransPubWrapper mixer with CDATASectionProxy."
    pass

class TransPubText(TextProxy, DOMProxy.TransPubCharacterData):
    "A TransPubWrapper mixer with TextProxy."
    pass

class TransPubComment(CommentProxy, DOMProxy.TransPubCharacterData):
    "A TransPubWrapper mixer with CommentProxy."
    pass

class TransPubProcessingInstruction(DOMProxy.ProcessingInstructionProxy,
                                    TransPubNode):
    "A TransPubWrapper mixer with ProcessingInstructionProxy."
    pass

class TransPubAttr(DOMProxy.AttrProxy, TransPubNode):
    "A TransPubWrapper mixer with AttrProxy."
    pass

class TransPubDocument(DOMProxy.DocumentProxy, TransPubNode):
    """
    A TransPubWrapper mixer with DocumentProxy.
    Provides and protects the implementation attribute.
    """
    
    implementation = theDOMImplementation

    #block set of implementation, since we don't proxy it the same
    def __setattr__(self, name, value):
        if name == "implementation":
            raise xml.dom.NoModificationAllowedErr()
        # wacky ExtensionClass inheritance
        TransPubDocument.inheritedAttribute('__setattr__')(self, name, value)

# DOM extended interfaces

class TransPubEntityReference(DOMProxy.EntityReferenceProxy, TransPubNode):
    "A TransPubWrapper mixer with EntityReferenceProxy."
    pass

class TransPubEntity(DOMProxy.EntityProxy, TransPubNode):
    "A TransPubWrapper mixer with EntityProxy."
    pass

class TransPubNotation(DOMProxy.NotationProxy, TransPubNode):
    "A TransPubWrapper mixer with NotationProxy."
    pass

class TransPubDocumentType(DOMProxy.DocumentTypeProxy, TransPubNode):
    "A TransPubWrapper mixer with DocumentTypeProxy."
    pass




##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"Test some ODB interactions."

import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase

ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML import ParsedXML

class ODBTestCase(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        self.document = ParsedXML.ParsedXML('foo')

    def shotgunTpStuff(self, node):
        "assert that the Tp* functions work"
        children = node.childNodes
        tpVals = node.tpValues()

        for i in range(children.length):
            assert children.item(i).isSameNode(tpVals[i])
            
        for i in range(children.length):
            assert children.item(i).tpURL() == str(i)
            assert children.item(i).tpId() == str(i)

        for node in tpVals:
            self.shotgunTpStuff(node)
        
    def testTpStuff(self):
        "check that the Tp* functions work"
        self.document.documentElement.appendChild(
            self.document.createElement('zero'))
        self.document.documentElement.appendChild(
            self.document.createElement('one'))
        self.document.documentElement.firstChild.appendChild(
            self.document.createElement('zero-zero'))
        self.document.documentElement.firstChild.appendChild(
            self.document.createElement('zero-one'))

        self.shotgunTpStuff(self.document.documentElement)        

if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(ODBTestCase))
        return suite



import urllib

def add_and_edit(self, id, REQUEST, screen='manage_main'):
    """Helper function to point to the object's management screen if
    'Add and Edit' button is pressed.
    id -- id of the object we just added
    """
    if REQUEST is None:
        return
    try:
        u = self.DestinationURL()
    except:
        u = REQUEST['URL1']
    if REQUEST.has_key('submit_edit'):
        u = "%s/%s" % (u, urllib.quote(id))
        REQUEST.RESPONSE.redirect(u+'/'+screen)
    else:
        REQUEST.RESPONSE.redirect(u+'/manage_main')



##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

from Base import *
import CoreLvl1
_DOMImplCase = CoreLvl1.DOMImplementationReadTestCase
_NodeTestCaseBase = CoreLvl1.NodeWriteTestCaseBase
del CoreLvl1

import sys
import string
import xml.dom
from xml.dom import Node

# --- DOMIMplementation

class DOMImplementationReadTestCase(TestCaseBase):

    TEST_NAMESPACE = TEST_NAMESPACE
    TEST_PREFIX = 'aprefix'
    TEST_LOCAL_NAME = 'somelocalname'
    TEST_QUALIFIED_NAME = '%s:%s' % (TEST_PREFIX, TEST_LOCAL_NAME)

    def setUp(self):
        # self.implementation is already set.
        pass

    def checkCreateDocument(self):
        # Non namespace Document
        newDoc = self.implementation.createDocument(None, self.TEST_LOCAL_NAME,
            None)

        checkAttribute(newDoc, 'nodeType', Node.DOCUMENT_NODE)
        checkAttribute(newDoc, 'doctype', None)
        checkAttributeNot(newDoc, 'documentElement', None)
        assert newDoc.implementation is self.implementation, (
            'Created Document has different implementation.')
        checkAttribute(newDoc.documentElement, 'namespaceURI', None)
        checkAttribute(newDoc.documentElement, 'tagName', self.TEST_LOCAL_NAME)

    def checkCreateDocumentWithNamespace(self):
        newDoc = self.implementation.createDocument(self.TEST_NAMESPACE,
            self.TEST_QUALIFIED_NAME, None)

        checkAttribute(newDoc.documentElement, 'namespaceURI', 
            self.TEST_NAMESPACE)
        checkAttribute(newDoc.documentElement, 'tagName', 
            self.TEST_QUALIFIED_NAME)

    def checkCreateDocumentWithDocumentType(self):
        docType = self.implementation.createDocumentType(
            self.TEST_QUALIFIED_NAME, 'uri:public', 'uri:system')
        newDoc = self.implementation.createDocument(None, self.TEST_LOCAL_NAME,
            docType)

        checkAttributeSameNode(newDoc, 'doctype', docType)

    def checkCreateDocumentIllegalCharacterErr(self):
        try:
            self.implementation.createDocument(self.TEST_NAMESPACE, 
                '4_prefix:5_illegal', None)
        except xml.dom.InvalidCharacterErr:
            pass
        else:
            assert 0, "Created Document with illegal qualified name."

    def checkCreateDocumentMalformedQA(self):
        try:
            self.implementation.createDocument(self.TEST_NAMESPACE,
                'malformed:qualfied:name', None)
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created document with a malformed qualified name."
        
        try:
            self.implementation.createDocument(self.TEST_NAMESPACE,
                ':malformed_qn', None)
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created document with a malformed qualified name."

    def checkCreateDocumentWithPrefixNoNamespace(self):
        try:
            self.implementation.createDocument(None, self.TEST_QUALIFIED_NAME,
                None)
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created document with a prefix but no namespace."

    def checkCreateDocumentWithXMLPrefixWrongNamespace(self):
        try:
            self.implementation.createDocument(self.TEST_NAMESPACE, 'xml:nope',
                None)
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, (
                "Created document with a 'xml' prefix but not the XML "
                "namespace.")

    def checkCreateDocumentWithUsedDocType(self):
        docType = self.implementation.createDocumentType(
            self.TEST_QUALIFIED_NAME, 'uri:public', 'uri:system')
        self.implementation.createDocument(None, self.TEST_LOCAL_NAME,
            docType)
        try:
            self.implementation.createDocument(None, self.TEST_LOCAL_NAME,
                docType)
        except xml.dom.WrongDocumentErr:
            pass
        else:
            assert 0, (
                "Used a doctype that was already in use to create a new "
                "Document.")

    def checkCreateDocumentType(self):
        docType = self.implementation.createDocumentType(
            self.TEST_QUALIFIED_NAME, 'uri:public', 'uri:system')

        checkAttribute(docType, 'ownerDocument', None)
        checkAttribute(docType, 'name', self.TEST_QUALIFIED_NAME)
        checkAttribute(docType, 'publicId', 'uri:public')
        checkAttribute(docType, 'systemId', 'uri:system')

        checkLength(docType.entities, 0)
        checkLength(docType.notations, 0)

    def checkCreateDocumentTypeIllegalCharacterErr(self):
        try:
            self.implementation.createDocumentType('4_prefix:5_illegal',
                'uri:public', 'uri:system')
        except xml.dom.InvalidCharacterErr:
            pass
        else:
            assert 0, "Created Document Type with illegal qualified name."

    def checkCreateDocumentTypeMalformedQA(self):
        try:
            self.implementation.createDocumentType('malformed:qualfied:name',
                'uri:public', 'uri:system')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created Document Type with a malformed qualified name."

        try:
            self.implementation.createDocumentType(':malformed_qn',
                'uri:public', 'uri:system')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created Document Type with a malformed qualified name."


# --- Node

class NodeReadTestCaseBase(TestCaseBase):
    
    TEST_NAMESPACE = TEST_NAMESPACE
    TEST_PREFIX = 'aprefix'
    TEST_LOCAL_NAME = 'somelocalname'
    TEST_QUALIFIED_NAME = '%s:%s' % (TEST_PREFIX, TEST_LOCAL_NAME)

    def checkLocalName(self):
        if self.node.nodeType in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            expect = self.TEST_LOCAL_NAME
            noNS = self.nodeNoNS
        else:
            expect = None
            noNS = None

        checkAttribute(self.node, 'localName', expect)
        checkReadOnly(self.node, 'localName')

        if noNS:
            checkAttribute(noNS, 'localName', None)

    def checkNameSpaceURI(self):
        if self.node.nodeType in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            expect = self.TEST_NAMESPACE
            noNS = self.nodeNoNS
        else:
            expect = None
            noNS = None

        checkAttribute(self.node, 'namespaceURI', expect)
        checkReadOnly(self.node, 'namespaceURI')

        if noNS:
            checkAttribute(noNS, 'namespaceURI', None)

    def checkOwnerDocument(self):
        if self.node.nodeType in (Node.DOCUMENT_NODE,
                Node.DOCUMENT_TYPE_NODE):
            expect = None

        else:
            expect = self.document

        checkAttributeSameNode(self.node, 'ownerDocument', expect)
        checkReadOnly(self.node, 'ownerDocument')

    def checkPrefix(self):
        if self.node.nodeType in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            expect = self.TEST_PREFIX
            noNS = self.nodeNoNS
        else:
            expect = None
            noNS = None

        checkAttribute(self.node, 'prefix', expect)

        if noNS:
            checkAttribute(noNS, 'prefix', None)

    def hasAttributes(self):
        assert not self.node.hasAttributes(), \
               "hasAttributes returned 'true' when 'false' was expected."

    featureMatrix = _DOMImplCase.featureMatrix

    def checkIsSupported(self):
        for feature, level in self.featureMatrix:
            result1 = self.node.isSupported(feature, level)
            result2 = self.node.isSupported(string.upper(feature), level)
            result3 = self.node.isSupported(string.lower(feature), level)

            expect = ((feature, level) in self.supportedFeatures)

            assert result1 == result2 and result1 == result3, (
                "Different results from different case feature string.")

            assert (result1 and 1 or 0) == expect, (
                "Test for %s, version %s should have returned %s, "
                "returned %s" % (repr(feature), repr(level), repr(expect),
                    repr(result1)))


class NodeWriteTestCaseBase(TestCaseBase):
    
    TEST_NAMESPACE = NodeReadTestCaseBase.TEST_NAMESPACE
    TEST_PREFIX = NodeReadTestCaseBase.TEST_PREFIX
    TEST_LOCAL_NAME = NodeReadTestCaseBase.TEST_LOCAL_NAME
    TEST_QUALIFIED_NAME = NodeReadTestCaseBase.TEST_QUALIFIED_NAME

    readOnlyNodeList = _NodeTestCaseBase.readOnlyNodeList

    def checkPrefix(self):
        # Nodetypes that are read-only.
        if self.node.nodeType in self.readOnlyNodeList:
            checkReadOnly(self.node, 'nodeValue')
            return

        # Nodetypes that should ignore changes.
        if self.node.nodeType not in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            self.node.prefix = "Ignore_this"
            checkAttribute(self.node, "prefix", None)
            return

        # All other node types
        self.node.prefix = 'foo'
        checkAttribute(self.node, 'prefix', 'foo')
        checkAttribute(self.node, 'nodeName', 'foo:%s' % self.TEST_LOCAL_NAME)
        
        # Changing the prefix also changes other attributes
        newQualifiedName = '%s:%s' % ('foo', self.TEST_LOCAL_NAME)
        checkAttribute(self.node, 'nodeName', newQualifiedName)
        if self.node.nodeType == Node.ATTRIBUTE_NODE:
            checkAttribute(self.node, 'name', newQualifiedName)
        else:
            checkAttribute(self.node, 'tagName', newQualifiedName)

    def checkPrefixInvalidCharacter(self):
        if self.node.nodeType in self.readOnlyNodeList:
            return

        if self.node.nodeType not in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            return

        try:
            self.node.prefix = '5_illegal_prefix'
        except xml.dom.InvalidCharacterErr:
            pass
        else:
            assert 0, "Setting of illegal prefix succeeded."
        
    def checkPrefixMalformedPrefix(self):
        if self.node.nodeType in self.readOnlyNodeList:
            return

        if self.node.nodeType not in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            return

        try:
            self.node.prefix = ':prefix_with_colon'
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Setting of malformed prefix with ':' succeeded."
        
    def checkPrefixNamespaceErr(self):
        if self.node.nodeType in self.readOnlyNodeList:
            return

        if self.node.nodeType not in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            return

        try:
            self.nodeNoNS.prefix = 'foo'
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, 'Changing prefix on Node without namespace succeeded.'

    def checkPrefixXMLNamespace(self):
        if self.node.nodeType in self.readOnlyNodeList:
            return

        if self.node.nodeType not in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            return

        try:
            self.node.prefix = 'xml'
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, ("Changing prefix to 'xml' on Node without "
                       "W3C XML namespace succeeded.")

    def checkPrefixXMLNSNamespace(self):
        if self.node.nodeType in self.readOnlyNodeList:
            return

        if self.node.nodeType not in (Node.ATTRIBUTE_NODE,
                Node.ELEMENT_NODE):
            return

        if self.node.nodeType == Node.ATTRIBUTE_NODE:
            try:
                self.node.prefix = 'xmlns'
            except xml.dom.NamespaceErr:
                pass
            else:
                assert 0, ("Changing prefix to 'xmlns' on Attribute without "
                           "W3C XML Namespaces namespace succeeded.")


# --- Document

class DocumentReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        self.node = self.createDocumentNS()

    def checkImportNode(self):
        foreignDoc = self.implementation.createDocument(None, 'foo', None)
        try:
            foreignDoc.importNode(self.document, 0)
        except xml.dom.NotSupportedErr:
            pass
        else:
            assert 0, "Was allowed to import a Document Node."


class DocumentWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        self.node = self.createDocumentNS()

    def checkGetElementsByTagNameNS(self):
        doc = self.document

        elements = {}
        names = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i']
        for c in names:
            elements[c + '1'] = doc.createElementNS('uri:1', 'one:' + c)
            elements[c + '2'] = doc.createElementNS('uri:2', 'two:' + c)
        # set up simple tree
        elements['a1'].appendChild(elements['a2'])
        elements['a1'].appendChild(elements['b1'])
        elements['a1'].appendChild(elements['d1'])
        elements['b1'].appendChild(elements['b2'])
        elements['b1'].appendChild(elements['c1'])
        elements['c1'].appendChild(elements['c2'])
        elements['d1'].appendChild(elements['d2'])
        elements['d1'].appendChild(elements['e1'])
        elements['d1'].appendChild(elements['h1'])
        elements['e1'].appendChild(elements['e2'])
        elements['e1'].appendChild(elements['f1'])
        elements['e1'].appendChild(elements['g1'])
        elements['f1'].appendChild(elements['f2'])
        elements['g1'].appendChild(elements['g2'])
        elements['h1'].appendChild(elements['h2'])
        elements['h1'].appendChild(elements['i1'])
        elements['i1'].appendChild(elements['i2'])

        doc.documentElement.appendChild(elements['a1'])
        
        # now test

        # find all elements in the right order
        result = doc.getElementsByTagNameNS('*', '*')
        allnames = []
        add = allnames.append
        for name in names:
            add(name)
            add(name)
        assert len(result) == len(allnames) + 1
        for name, element in map(None, allnames, result[1:]):
            assert name == element.localName

        # find all elements in one namespace in the right order
        result = doc.getElementsByTagNameNS('uri:1', '*')
        assert len(result) == len(names)
        for name, element in map(None, names, result):
            assert name == element.localName

        # find single element, top
        result = doc.getElementsByTagNameNS('uri:1', 'a')
        assert len(result) == 1
        assert result[0].tagName == 'one:a'

        # find single element somewhere in tree
        result = doc.getElementsByTagNameNS('uri:2', 'h')
        assert len(result) == 1
        assert result[0].tagName == 'two:h'

        # find elements in the tree from all namespaces
        result = doc.getElementsByTagNameNS('*', 'f')
        assert len(result) == 2
        assert result[0].tagName == 'one:f'
        assert result[1].tagName == 'two:f'

    def checkNormalize(self):
        doc = self.document
        docEl = doc.documentElement

        # First build a tree with adjacent and empty Text nodes, intermingled
        # with other Nodes.
        docEl.appendChild(doc.createTextNode('This'))
        docEl.appendChild(doc.createTextNode(''))
        docEl.appendChild(doc.createTextNode(' is'))
        docEl.appendChild(doc.createTextNode(' '))
        docEl.appendChild(doc.createTextNode('a test.'))

        docEl.appendChild(doc.createCDATASection("Don't merge"))
        docEl.appendChild(doc.createTextNode('1'))

        docEl.appendChild(doc.createComment('foo'))
        docEl.appendChild(doc.createTextNode('2'))

        subEl = doc.createElement('baz')
        attr = doc.createAttribute('eggs')
        subEl.setAttributeNode(attr)
        docEl.appendChild(subEl)

        subEl.appendChild(doc.createTextNode('To '))
        subEl.appendChild(doc.createTextNode(''))
        subEl.appendChild(doc.createTextNode('be or'))
        
        subEl.appendChild(doc.createEntityReference('amp'))

        subEl.appendChild(doc.createTextNode('not '))
        subEl.appendChild(doc.createTextNode('to'))
        subEl.appendChild(doc.createTextNode(' be'))

        attr.appendChild(doc.createTextNode('Spanish '))
        attr.appendChild(doc.createTextNode('Inquisition'))

        docEl.appendChild(doc.createTextNode('3'))

        docEl.appendChild(doc.createProcessingInstruction('bar', 'spam'))
        docEl.appendChild(doc.createTextNode('4'))

        # Now test
        doc.normalize()

        checkAttribute(docEl.childNodes, 'length', 9)
        checkAttribute(subEl.childNodes, 'length', 3)
        checkAttribute(attr.childNodes, 'length', 1)

        checkAttribute(docEl.childNodes[0], 'nodeType', Node.TEXT_NODE)
        checkAttribute(docEl.childNodes[0], 'data', 'This is a test.')

        checkAttribute(docEl.childNodes[1], 'nodeType', 
            Node.CDATA_SECTION_NODE)
        checkAttribute(docEl.childNodes[2], 'nodeType', Node.TEXT_NODE)
        checkAttribute(docEl.childNodes[3], 'nodeType', 
            Node.COMMENT_NODE)
        checkAttribute(docEl.childNodes[4], 'nodeType', Node.TEXT_NODE)
        checkAttribute(docEl.childNodes[5], 'nodeType',
            Node.ELEMENT_NODE)
        checkAttribute(docEl.childNodes[6], 'nodeType', Node.TEXT_NODE)
        checkAttribute(docEl.childNodes[7], 'nodeType',
            Node.PROCESSING_INSTRUCTION_NODE)
        checkAttribute(docEl.childNodes[8], 'nodeType', Node.TEXT_NODE)

        checkAttribute(subEl.childNodes[0], 'nodeType', Node.TEXT_NODE)
        checkAttribute(subEl.childNodes[1], 'nodeType', 
            Node.ENTITY_REFERENCE_NODE)
        checkAttribute(subEl.childNodes[2], 'nodeType', Node.TEXT_NODE)

        checkAttribute(subEl.childNodes[0], 'data', 'To be or')
        checkAttribute(subEl.childNodes[2], 'data', 'not to be')

        checkAttribute(attr.childNodes[0], 'data', 'Spanish Inquisition')

    def checkCreateAttributeNS(self):
        attr = self.document.createAttributeNS(self.TEST_NAMESPACE,
            self.TEST_QUALIFIED_NAME)

        checkAttribute(attr, 'nodeType', Node.ATTRIBUTE_NODE)
        checkAttribute(attr, 'name', self.TEST_QUALIFIED_NAME)
        checkAttribute(attr, 'localName', self.TEST_LOCAL_NAME)
        checkAttribute(attr, 'prefix', self.TEST_PREFIX)
        checkAttribute(attr, 'namespaceURI', self.TEST_NAMESPACE)
        checkAttribute(attr, 'value', '')
        checkAttributeSameNode(attr, 'ownerDocument', self.document)

    def checkCreateAttributeNSNoPrefix(self):
        attr = self.document.createAttributeNS(self.TEST_NAMESPACE,
            self.TEST_LOCAL_NAME)

        checkAttribute(attr, 'name', self.TEST_LOCAL_NAME)
        checkAttribute(attr, 'localName', self.TEST_LOCAL_NAME)
        checkAttribute(attr, 'prefix', None)
        checkAttribute(attr, 'namespaceURI', self.TEST_NAMESPACE)

    def checkCreateAttributeNSIllegalCharacter(self):
        try:
            self.document.createAttributeNS(self.TEST_NAMESPACE, 
                '4_prefix:5_illegal')
        except xml.dom.InvalidCharacterErr:
            pass
        else:
            assert 0, "Created Attribute with illegal qualified name."

    def checkCreateAttributeNSMalformedQA(self):
        try:
            self.document.createAttributeNS(self.TEST_NAMESPACE, 
                'malformed:qualfied:name')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created Attribute with a malformed qualified name."

        try:
            self.document.createAttributeNS(self.TEST_NAMESPACE, 
                ':malformed_qn')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created Attribute with a malformed qualified name."

    def checkCreateAttributeNSPrefixNoNamespace(self):
        try:
            self.document.createAttributeNS(None,
                self.TEST_QUALIFIED_NAME)
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created Attribute with a prefix but no namespace."

    def checkCreateAttributeNSXMLNamespace(self):
        try:
            self.document.createAttributeNS(self.TEST_NAMESPACE,
                'xml:nope')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, (
                "Created Attribute with a 'xml' prefix but not the XML "
                "namespace.")

    def checkCreateAttributeNSXMLNamespace(self):
        try:
            self.document.createAttributeNS(self.TEST_NAMESPACE,
                'xmlns:nope')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, (
                "Created Attribute with a 'xmlns' prefix but not the XML "
                "Namespaces namespace.")

    def checkCreateElementNS(self):
        el = self.document.createElementNS(self.TEST_NAMESPACE,
            self.TEST_QUALIFIED_NAME)

        checkAttribute(el, 'nodeType', Node.ELEMENT_NODE)
        checkAttribute(el, 'tagName', self.TEST_QUALIFIED_NAME)
        checkAttribute(el, 'localName', self.TEST_LOCAL_NAME)
        checkAttribute(el, 'prefix', self.TEST_PREFIX)
        checkAttribute(el, 'namespaceURI', self.TEST_NAMESPACE)
        checkAttributeSameNode(el, 'ownerDocument', self.document)

    def checkCreateElementNSNoPrefix(self):
        el = self.document.createElementNS(self.TEST_NAMESPACE,
            self.TEST_LOCAL_NAME)

        checkAttribute(el, 'tagName', self.TEST_LOCAL_NAME)
        checkAttribute(el, 'localName', self.TEST_LOCAL_NAME)
        checkAttribute(el, 'prefix', None)
        checkAttribute(el, 'namespaceURI', self.TEST_NAMESPACE)

    def checkCreateElementNSIllegalCharacter(self):
        try:
            self.document.createElementNS(self.TEST_NAMESPACE, 
                '4_prefix:5_illegal')
        except xml.dom.InvalidCharacterErr:
            pass
        else:
            assert 0, "Created Element with illegal qualified name."

    def checkCreateElementNSMalformedQA(self):
        try:
            self.document.createElementNS(self.TEST_NAMESPACE,
                'malformed:qualfied:name')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created Element with a malformed qualified name."
        try:
            self.document.createElementNS(self.TEST_NAMESPACE, ':malformed_qn')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created Element with a malformed qualified name."

    def checkCreateElementNSPrefixNoNamespace(self):
        try:
            self.document.createElementNS(None,
                self.TEST_QUALIFIED_NAME)
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Created Element with a prefix but no namespace."

    def checkCreateElementNSXMLNamespace(self):
        try:
            self.document.createElementNS(self.TEST_NAMESPACE,
                'xml:nope')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, (
                "Created Element with a 'xml' prefix but not the XML "
                "namespace.")


class GetElementByIdTestCase(TestCaseBase):

    PROLOGUE = """\
<!DOCTYPE idtest [
  <!ELEMENT idtest EMPTY>
  <!ATTLIST idtest
            id   ID   #IMPLIED
    >
]>
"""

    def setup(self):
        pass

    def checkWithoutId(self):
        doc = self.parse(self.PROLOGUE + "<idtest/>")
        self.assert_(doc.getElementById("foo") is None)

    def checkWithDifferentId(self):
        doc = self.parse(self.PROLOGUE + "<idtest id='bar'/>")
        self.assert_(doc.getElementById("foo") is None)

    def checkWithIdOnRootElement(self):
        doc = self.parse(self.PROLOGUE + "<idtest id='foo'/>")
        self.assert_(isSameNode(doc.getElementById("foo"),
                                doc.documentElement))

    def checkWithIdOnChildElement(self):
        doc = self.parse(self.PROLOGUE
                         + "<idtest><idtest id='foo'/></idtest>")
        self.assert_(isSameNode(doc.getElementById("foo"),
                                doc.documentElement.firstChild))


# --- Element

class ElementReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        doc = self.createDocument()
        self.element = self.nodeNoNS = self.elementNoNS = doc.createElement(
            self.TEST_LOCAL_NAME)
        self.elementNS = self.node = doc.createElementNS(
            self.TEST_NAMESPACE, self.TEST_QUALIFIED_NAME)

    def checkGetAttributeNS(self):
        assert self.element.getAttributeNS("uri:bugga", "ugga") == "", \
               "non-existant attribute should return ''"

    def checkGetAttributeNodeNS(self):
        assert self.element.getAttributeNodeNS("uri:bugga", "ugga") is None, \
               "non-existant attribute should return None"

    def checkCloneNode(self):
        el = self.element
        clone = el.cloneNode(0)

        assert not isSameNode(el, clone), "Clone is same Node as original."
        checkAttribute(clone, 'localName', el.localName)
        checkAttribute(clone, 'namespaceURI', el.namespaceURI)
        checkAttribute(clone, 'prefix', el.prefix)

    def checkImportNode(self):
        el = self.element
        doc = self.document

        el.appendChild(doc.createTextNode('A Text Node'))
        el.appendChild(doc.createComment('A Comment'))
        el.setAttribute('attr1', 'An attribute')
        el.setAttribute('attr2', 'Another attribute')
        
        foreignDoc = self.implementation.createDocument(None, 'foo', None)

        clone = foreignDoc.importNode(self.element, 0)
        deepClone = foreignDoc.importNode(self.element, 1)

        assert not isSameNode(el, clone), "Clone is same Node as original."
        assert not isSameNode(el, deepClone), "Clone is same Node as original."

        checkAttributeSameNode(clone, 'ownerDocument', foreignDoc)
        checkAttributeSameNode(deepClone, 'ownerDocument', foreignDoc)
        checkAttribute(clone, 'parentNode', None)
        checkAttribute(deepClone, 'parentNode', None)
        checkLength(clone.childNodes, 0)
        checkLength(deepClone.childNodes, el.childNodes.length)
        checkLength(clone.attributes, el.attributes.length)
        checkLength(deepClone.attributes, el.attributes.length)

        checkAttribute(clone, 'nodeName', el.nodeName)
        checkAttribute(deepClone, 'nodeName', el.nodeName)
        checkAttribute(clone, 'nodeType', el.nodeType)
        checkAttribute(deepClone, 'nodeType', el.nodeType)
        checkAttribute(clone, 'nodeValue', el.nodeValue)
        checkAttribute(deepClone, 'nodeValue', el.nodeValue)

        for i in range(deepClone.childNodes.length):
            checkAttribute(deepClone.childNodes.item(i), 'nodeType',
                el.childNodes.item(i).nodeType)
            checkAttributeSameNode(deepClone.childNodes.item(i),
                'ownerDocument', foreignDoc)
            if deepClone.childNodes.item(i).nodeType != Node.ELEMENT_NODE:
                checkAttribute(deepClone.childNodes.item(i), 'data',
                    el.childNodes.item(i).data)

        for i in range(deepClone.attributes.length):
            checkAttribute(clone.attributes.item(i), 'name',
                el.attributes.item(i).name)
            checkAttribute(deepClone.attributes.item(i), 'name',
                el.attributes.item(i).name)

            checkAttribute(clone.attributes.item(i), 'value',
                el.attributes.item(i).value)
            checkAttribute(deepClone.attributes.item(i), 'value',
                el.attributes.item(i).value)

            checkAttributeSameNode(clone.attributes.item(i), 'ownerElement',
                clone)
            checkAttributeSameNode(deepClone.attributes.item(i),
                'ownerElement', deepClone)


class ElementWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        doc = self.createDocument()
        self.element = self.nodeNoNS = self.elementNoNS = doc.createElement(
            self.TEST_LOCAL_NAME)
        self.elementNS = self.node = doc.createElementNS(self.TEST_NAMESPACE,
            self.TEST_QUALIFIED_NAME)

    def checkHasAttribute(self):
        assert not self.element.hasAttribute('ugga'), \
               "Test for non-exisiting attribute returned true."

        self.element.setAttribute("ugga", "foo")

        assert self.element.hasAttribute('ugga'), \
               "Test for exisiting attribute returned false."

    def checkSetAttributeNS(self):
        self.element.setAttributeNS("uri:bugga", "b:ugga", "foo")
        self.element.setAttributeNS("uri:bar", "bar:ugga", "baz")

        assert self.element.hasAttributeNS("uri:bar", "ugga"), \
               "Test for presence of created attribute returned false."
        value = self.element.getAttributeNS("uri:bugga", "ugga")
        assert value == 'foo', (
            "Incorrect attr value returned. Expected 'foo', got %s"
                % repr(value))
        node = self.element.getAttributeNodeNS("uri:bugga", "ugga")
        assert node.prefix == 'b', (
            "New Attr node has incorrect prefix, expected 'b', got "
            "%s" % repr(node.prefix))

    #"Note that because the DOM does no lexical checking, the empty string
    #will be treated as a real namespace URI in DOM Level 2 methods.
    #Applications must use the value null as the namespaceURI parameter for
    #methods if they wish to have no namespace." - the rec
    # The NamespaceURI must be a namespace name, which Namespaces in XML
    # states is a URI reference.  I'm assuming that "" is not a valid URI
    # reference, so we're not testing using that.
    def checkSetAttributeEmptyNS(self):
        self.element.setAttributeNS(None, "ugga", "foo")

        assert self.element.hasAttributeNS(None, "ugga"), \
               "Test for presence of created attribute returned false."
        value = self.element.getAttributeNS(None, "ugga")
        assert value == 'foo', (
            "Incorrect attr value returned. Expected 'foo', got %s"
                % repr(value))
        node = self.element.getAttributeNodeNS(None, "ugga")
        assert node.prefix == None, (
            "New Attr node has incorrect prefix, expected None, got "
            "%s" % repr(node.prefix))

    def checkSetAttributeNSDifferentPrefix(self):
        self.element.setAttributeNS("uri:bugga", "b:ugga", "foo")
        self.element.setAttributeNS("uri:bar", "bar:ugga", "baz")

        self.element.setAttributeNS("uri:bugga", "c:ugga", "new value")

        assert self.element.hasAttributeNS("uri:bar", "ugga"), (
               "Test for presence of created attribute returned false.")
        value = self.element.getAttributeNS("uri:bugga", "ugga")
        assert value == 'new value', (
               "Incorrect attr value returned. Expected 'new value', got "
               "%s" % repr(value))
        node = self.element.getAttributeNodeNS("uri:bugga", "ugga")
        assert node.prefix == 'c', (
            "New Attr node has incorrect prefix, expected 'c', got "
            "%s" % repr(node.prefix))

    def checkSetAttributeNSNoPrefix(self):
        self.element.setAttributeNS(TEST_NAMESPACE, 'foo', 'bar')

        assert self.element.hasAttributeNS(TEST_NAMESPACE, "foo"), (
               "Test for presence of created attribute returned false.")
        value = self.element.getAttributeNS(TEST_NAMESPACE, "foo")
        assert value == 'bar', (
               "Incorrect attr value returned. Expected 'bar', got "
               "%s" % repr(value))
        node = self.element.getAttributeNodeNS(TEST_NAMESPACE, "foo")
        checkAttribute(node, 'prefix', None)

    def checkSetAttributeNSXMLNSDeclaration(self):
        XMLNSNamespace = 'http://www.w3.org/2000/xmlns/'
        self.element.setAttributeNS(XMLNSNamespace, 'xmlns', TEST_NAMESPACE)

        assert self.element.hasAttributeNS(XMLNSNamespace, "xmlns"), (
               "Test for presence of created xmlns attribute returned false.")
        value = self.element.getAttributeNS(XMLNSNamespace, "xmlns")
        assert value == TEST_NAMESPACE, (
               "Incorrect attr value returned. Expected %s, got "
               "%s" % (`TEST_NAMESPACE`, `value`))
        node = self.element.getAttributeNodeNS(XMLNSNamespace, "xmlns")
        checkAttribute(node, 'prefix', None)

    def checkSetAttributeNSIllegalCharacter(self):
        try:
            self.element.setAttributeNS("uri:bugga", "5_b:ugga", "illegal")
        except xml.dom.InvalidCharacterErr:
            pass
        else:
            assert 0, "Was allowed to use an illegal attribute name."

    def checkSetAttributeNSMalformedQA(self):
        try:
            self.element.setAttributeNS("uri:bugga", 'malformed:qualfied:name',
                "malformed")
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Was allowed to use a malformed qualified name."
        try:
            self.element.setAttributeNS("uri:bugga", ':malformed_qn',
                "malformed")
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, "Was allowed to use a malformed qualified name."

    def checkSetAttributeNSPrefixNoNamespace(self):
        try:
            self.element.setAttributeNS(None, 'prefix:localName', 'Nono')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, (
                'Using a null namespace and a prefix for an '
                'attribute succeeded.')

    def checkSetAttributeNSXMLNamespace(self):
        try:
            self.element.setAttributeNS('uri:unknown', 'xml:localName', 'Nono')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, ("Using the prefix 'xml' for an attribute without "
                       "W3C XML namespace succeeded.")

    def checkSetAttributeNSXMLNSNamespace(self):
        try:
            self.element.setAttributeNS('uri:unknown', 'xmlns:localName', 'No')
        except xml.dom.NamespaceErr:
            pass
        else:
            assert 0, ("Using the prefix 'xmlns' for an attribute without "
                       "W3C XML Namespaces namespace succeeded.")

    def checkSetAttributeNodeNS(self):
        node1 = self.document.createAttributeNS("uri:bugga", "b:ugga")
        node1.value = 'foo'
        node2 = self.document.createAttributeNS("uri:bar", "bar:ugga")
        returnValue = self.element.setAttributeNode(node1)
        self.element.setAttributeNode(node2)

        assert returnValue is None, (
            "setAttributeNodeNS returned %s" % repr(returnValue))
        
        assert self.element.hasAttributeNS("uri:bar", "ugga"), \
               "Test for presence of created attribute returned false."

        value = self.element.getAttributeNS("uri:bugga", "ugga")
        assert value == 'foo', \
               ("Incorrect attr value returned. Expected 'foo', got %s"
                % repr(value))

        node = self.element.getAttributeNodeNS("uri:bugga", "ugga")
        assert isSameNode(node1, node), \
               "Incorrect node returned from getAttributeNodeNS."

    def checkSetAttributeNodeNSReplaceExisting(self):
        # Try a new node with same namespaceURI and localname, but differing
        # prefix. This should replace the existing node, returning it.
        node1 = self.document.createAttributeNS("uri:bugga", "b:ugga")
        self.element.setAttributeNode(node1)
        node2 = self.document.createAttributeNS("uri:bugga", "c:ugga")

        returnValue = self.element.setAttributeNodeNS(node2)
        if returnValue is None:
            assert 0, "setAttributeNodeNS did not replace original attribute"
        assert isSameNode(node1, returnValue), (
            "setAttributeNodeNS returned %s" % repr(returnValue))

    def checkSetAttributeNodeNSWrongDocument(self):
        foreignDoc = self.implementation.createDocument(None, 'foo', None)
        foreignAttr = foreignDoc.createAttributeNS('uri:spam', 'spam:eggs')
        try:
            self.element.setAttributeNodeNS(foreignAttr)
        except xml.dom.WrongDocumentErr:
            pass
        else:
            assert 0, "Was allowed to setAttributeNodeNS with a foreign Attr."

    def checkSetAttributeNodeNSAlreadyInUse(self):
        otherElement = self.document.createElement('foo')
        otherAttr = self.document.createAttributeNS('uri:spam', 'spam:eggs')
        otherElement.setAttributeNodeNS(otherAttr)
        try:
            self.element.setAttributeNodeNS(otherAttr)
        except xml.dom.InuseAttributeErr:
            pass
        else:
            assert 0, (
                "Was allowed to setAttributeNodeNS with an Attr already in "
                "use.")

    def checkRemoveAttributeNS(self):
        node1 = self.document.createAttributeNS("uri:bugga", "b:ugga")
        node2 = self.document.createAttributeNS("uri:bar", "bar:ugga")
        self.element.setAttributeNode(node1)
        self.element.setAttributeNode(node2)

        self.element.removeAttributeNS("uri:bugga", "ugga")
        
        assert not self.element.hasAttributeNS("uri:bugga", "ugga"), \
               "Test for presence of created attribute still returns true."

        assert self.element.hasAttributeNS("uri:bar", "ugga"), \
               "Test for presence of created attribute returned false."

    def checkGetElementsByTagNameNS(self):
        doc = self.document
        el = self.element

        elements = {}
        names = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i']
        for c in names:
            elements[c + '1'] = doc.createElementNS('uri:1', 'one:' + c)
            elements[c + '2'] = doc.createElementNS('uri:2', 'two:' + c)
        # set up simple tree
        elements['a1'].appendChild(elements['a2'])
        elements['a1'].appendChild(elements['b1'])
        elements['a1'].appendChild(elements['d1'])
        elements['b1'].appendChild(elements['b2'])
        elements['b1'].appendChild(elements['c1'])
        elements['c1'].appendChild(elements['c2'])
        elements['d1'].appendChild(elements['d2'])
        elements['d1'].appendChild(elements['e1'])
        elements['d1'].appendChild(elements['h1'])
        elements['e1'].appendChild(elements['e2'])
        elements['e1'].appendChild(elements['f1'])
        elements['e1'].appendChild(elements['g1'])
        elements['f1'].appendChild(elements['f2'])
        elements['g1'].appendChild(elements['g2'])
        elements['h1'].appendChild(elements['h2'])
        elements['h1'].appendChild(elements['i1'])
        elements['i1'].appendChild(elements['i2'])

        el.appendChild(elements['a1'])
        
        # now test

        # find all elements in the right order
        result = el.getElementsByTagNameNS('*', '*')
        allnames = []
        add = allnames.append
        for name in names:
            add(name)
            add(name)
        assert len(result) == len(allnames)
        for name, element in map(None, allnames, result):
            assert name == element.localName

        # find all elements in one namespace in the right order
        result = el.getElementsByTagNameNS('uri:1', '*')
        assert len(result) == len(names)
        for name, element in map(None, names, result):
            assert name == element.localName

        # find single element, top
        result = el.getElementsByTagNameNS('uri:1', 'a')
        assert len(result) == 1
        assert result[0].tagName == 'one:a'

        # find single element somewhere in tree
        result = el.getElementsByTagNameNS('uri:2', 'h')
        assert len(result) == 1
        assert result[0].tagName == 'two:h'

        # find elements in the tree from all namespaces
        result = el.getElementsByTagNameNS('*', 'f')
        assert len(result) == 2
        assert result[0].tagName == 'one:f'
        assert result[1].tagName == 'two:f'

    def checkNormalize(self):
        doc = self.document
        el = self.element

        # Build test nodes
        el1 = el.appendChild(doc.createElement('e1'))
        el2 = el.appendChild(doc.createElement('e2'))
        
        el1.appendChild(doc.createTextNode('foo'))
        el1.appendChild(doc.createTextNode('bar'))

        el2.appendChild(doc.createTextNode('spam'))
        el2.appendChild(doc.createTextNode('eggs'))

        # Now test
        # Only normalize on element, sibling should be unaffected
        el1.normalize()

        checkAttribute(el1.childNodes, 'length', 1)
        checkAttribute(el2.childNodes, 'length', 2)

        checkAttribute(el1.childNodes[0], 'data', 'foobar')

        # Now normalize the whole test tree
        el.normalize()

        checkAttribute(el1.childNodes, 'length', 1)
        checkAttribute(el2.childNodes, 'length', 1)

        checkAttribute(el1.childNodes[0], 'data', 'foobar')
        checkAttribute(el2.childNodes[0], 'data', 'spameggs')


# --- CharacterData

class CharacterDataReadTestCaseBase(NodeReadTestCaseBase):

    def checkImportNode(self):
        foreignDoc = self.implementation.createDocument(None, 'foo', None)

        clone = foreignDoc.importNode(self.chardata, 0)
        deepClone = foreignDoc.importNode(self.chardata, 1)

        assert not isSameNode(self.chardata, clone), (
            "Clone is same as original.")
        assert not isSameNode(self.chardata, deepClone), (
            "Clone is same as original.")
        
        checkAttributeSameNode(clone, 'ownerDocument', foreignDoc)
        checkAttributeSameNode(deepClone, 'ownerDocument', foreignDoc)
        checkAttribute(clone, 'parentNode', None)
        checkAttribute(deepClone, 'parentNode', None)
        checkAttribute(clone, 'nodeType', self.chardata.nodeType)
        checkAttribute(deepClone, 'nodeType', self.chardata.nodeType)
        checkAttribute(clone, 'data', self.chardata.data)
        checkAttribute(deepClone, 'data', self.chardata.data)
        checkLength(clone.childNodes, 0)
        checkLength(deepClone.childNodes, 0)


class CharacterDataWriteTestCaseBase(NodeWriteTestCaseBase):

    pass


# --- Comment

class CommentReadTestCase(CharacterDataReadTestCaseBase):

    def setUp(self):
        self.chardata = self.node = self.createDocument().createComment("com")


class CommentWriteTestCase(CharacterDataWriteTestCaseBase):
    
    def setUp(self):
        self.chardata = self.node = self.createDocument().createComment("com")


# --- Text

class TextReadTestCase(CharacterDataReadTestCaseBase):

    def setUp(self):
        self.chardata = self.node = self.createDocument().createTextNode("com")
        self.expectedType = Node.TEXT_NODE


class TextWriteTestCase(CharacterDataWriteTestCaseBase):

    def setUp(self):
        self.chardata = self.node = self.createDocument().createTextNode("com")
        self.expectedType = Node.TEXT_NODE


# --- Attr

class AttrReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        self.attr = self.node = self.createDocument().createAttributeNS(
            self.TEST_NAMESPACE, self.TEST_QUALIFIED_NAME)
        self.attrNoNS = self.nodeNoNS = self.document.createAttribute(
            self.TEST_LOCAL_NAME)

    def checkCloneNode(self):
        attr = self.attr
        clone = attr.cloneNode(0)

        assert not isSameNode(attr, clone), "Clone is same Node as original."
        checkAttribute(clone, 'localName', attr.localName)
        checkAttribute(clone, 'namespaceURI', attr.namespaceURI)
        checkAttribute(clone, 'prefix', attr.prefix)

        # make sure the cloned attr isn't sharing data with the original
        newPrefix = 'foo'
        newQname = '%s:%s' % (newPrefix, self.TEST_LOCAL_NAME)
        oldPrefix = self.attr.prefix
        oldQname = self.attr.name
        self.attr.prefix = newPrefix
        checkAttribute(clone, 'prefix', oldPrefix)
        checkAttribute(clone, 'name', oldQname)
        
    def checkImportNode(self):
        foreignDoc = self.implementation.createDocument(None, 'foo', None)

        clone = foreignDoc.importNode(self.attr, 0)
        deepClone = foreignDoc.importNode(self.attr, 1)

        assert not isSameNode(self.attr, clone), "Clone is same as original."
        assert not isSameNode(self.attr, deepClone), (
            "Clone is same as original.")
        
        checkAttributeSameNode(clone, 'ownerDocument', foreignDoc)
        checkAttributeSameNode(deepClone, 'ownerDocument', foreignDoc)
        checkAttribute(clone, 'parentNode', None)
        checkAttribute(deepClone, 'parentNode', None)
        checkAttribute(clone, 'nodeType', self.attr.nodeType)
        checkAttribute(deepClone, 'nodeType', self.attr.nodeType)
        checkAttribute(clone, 'name', self.attr.name)
        checkAttribute(deepClone, 'name', self.attr.name)
        checkAttribute(clone, 'value', self.attr.value)
        checkAttribute(deepClone, 'value', self.attr.value)
        checkAttribute(clone, 'specified', 1)
        checkAttribute(deepClone, 'specified', 1)
        checkAttribute(clone, 'nodeName', self.attr.nodeName)
        checkAttribute(deepClone, 'nodeName', self.attr.nodeName)
        checkAttribute(clone, 'nodeValue', self.attr.nodeValue)
        checkAttribute(deepClone, 'nodeValue', self.attr.nodeValue)

        checkLength(clone.childNodes, 1) # Subtree models value
        checkAttribute(clone.firstChild, 'nodeType', Node.TEXT_NODE)
        checkAttribute(clone.firstChild, 'data', self.attr.value)
        checkLength(deepClone.childNodes, 1)
        checkAttribute(deepClone.firstChild, 'nodeType', Node.TEXT_NODE)
        checkAttribute(deepClone.firstChild, 'data', self.attr.value)


class AttrWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        self.attr = self.node = self.createDocument().createAttributeNS(
            self.TEST_NAMESPACE, self.TEST_QUALIFIED_NAME)
        self.attrNoNS = self.nodeNoNS = self.document.createAttribute(
            self.TEST_LOCAL_NAME)

    def checkAttrNodePrefixMultipleRefs(self):
        "changing the attribute prefix should change the name of other refs"
        # qualified name with different prefix
        newPrefix = 'foo'
        newQname = '%s:%s' % (newPrefix, self.TEST_LOCAL_NAME)
        
        self.attr.value = 'spam'
        self.document.documentElement.setAttributeNode(self.attr)
        attr2 = self.document.documentElement.getAttributeNodeNS(
            self.TEST_NAMESPACE, self.TEST_LOCAL_NAME)

        self.attr.prefix = newPrefix

        attr3 = self.document.documentElement.getAttributeNodeNS(
            self.TEST_NAMESPACE, self.TEST_LOCAL_NAME)
        # orig attr
        checkAttribute(self.attr, 'nodeName', newQname)
        # attr gotten before set
        checkAttribute(attr2, 'nodeName', newQname)
        # attr gotten after set
        checkAttribute(attr3, 'nodeName', newQname)

    def checkElementAttrPrefixMultipleRefs(self):
        "changing the attribute prefix should change the name of other refs"
        # qualified name with different prefix
        newPrefix = 'foo'
        newQname = '%s:%s' % (newPrefix, self.TEST_LOCAL_NAME)

        self.attr.value = 'spam'
        self.document.documentElement.setAttributeNode(self.attr)
        attr2 = self.document.documentElement.getAttributeNodeNS(
            self.TEST_NAMESPACE, self.TEST_LOCAL_NAME)

        # change the prefix and value of the attr.  Since we're giving the same
        # localname and namespaceURI, the existing attr should be changed.
        self.document.documentElement.setAttributeNS(self.TEST_NAMESPACE,
                                                     newQname, 'eggs')

        attr3 = self.document.documentElement.getAttributeNodeNS(
            self.TEST_NAMESPACE, self.TEST_LOCAL_NAME)
        
        # element attr
        assert self.document.documentElement.getAttributeNS(
            self.TEST_NAMESPACE, self.TEST_LOCAL_NAME) == 'eggs', (
            "changing prefix and value on attr didn't change original attr")
        # orig attr
        checkAttribute(self.attr, 'nodeName', newQname)
        # attr gotten before set
        checkAttribute(attr2, 'nodeName', newQname)
        # attr gotten after set
        checkAttribute(attr3, 'nodeName', newQname)


# --- Default attributes

class DefaultAttrTestCase(TestCaseBase):

    def setUp(self):
        self.document = self.parse("""
            <!DOCTYPE doc [
                <!ELEMENT doc EMPTY>
                <!ATTLIST doc foo CDATA "bar">
            ]>
            <doc xmlns="%s"/>
        """ % TEST_NAMESPACE)

    def checkCreateElementNS(self):
        el = self.document.createElementNS(TEST_NAMESPACE, 'doc')

        assert el.hasAttribute('foo'), (
            'Newly created Element Node should have default attribute.')
        assert el.getAttribute('foo') == 'bar', (
            "Wrong value of default attribute found, expected 'bar', "
            "found %s" % el.getAttribute('foo'))
        checkAttribute(el.getAttributeNode('foo'), 'specified', 0)

    def checkImportNode(self):
        attr = self.document.documentElement.getAttributeNode('foo')
        newDoc = self.implementation.createDocument(None, 'baz', None)

        importedAttr = newDoc.importNode(attr, 0)
        checkAttribute(importedAttr, 'specified', 1)

    def checkImportNodeFromDefault(self):
        newDoc = self.implementation.createDocument(None, 'baz', None)
        
        el = newDoc.importNode(self.document.documentElement, 0)

        assert not el.hasAttribute('foo'), (
            "Default attribute retained when importing into document that "
            "doesn't specify the default attribute.")

    def checkImportNodeToDefault(self):
        newDoc = self.implementation.createDocument(None, 'baz', None)
        newEl = newDoc.createElementNS(TEST_NAMESPACE, 'doc')
        
        el = self.document.importNode(newEl, 0)

        assert el.hasAttribute('foo'), (
            'Imported Element Node should have default attribute.')
        assert el.getAttribute('foo') == 'bar', (
            "Wrong value of default attribute found, expected 'bar', "
            "found %s" % repr(el.getAttribute('foo')))
        checkAttribute(el.getAttributeNode('foo'), 'specified', 0)


class DefaultAttrWithPrefixTestCase(TestCaseBase):
    pass # XXX removed until we can verify default attributes in DOM spec
##     def setUp(self):
##         self.document = self.parse("""
##             <!DOCTYPE prefix:doc [
##                 <!ELEMENT prefix:doc EMPTY>
##                 <!ATTLIST prefix:doc prefix:foo CDATA "bar">
##             ]>
##             <prefix:doc xmlns:prefix="%s"/>
##         """ % TEST_NAMESPACE)

##     def checkHasAttributeNS(self):
##         el = self.document.documentElement

##         assert el.hasAttributeNS(TEST_NAMESPACE, 'foo'), (
##             'Default attribute not found.')

##     def checkGetAttributeNS(self):
##         el = self.document.documentElement
        
##         assert el.getAttributeNS(TEST_NAMESPACE, 'foo') == 'bar', (
##             "Wrong value of default attribute found, expected 'bar', "
##             "found %s" % repr(el.getAttributeNS(TEST_NAMESPACE, 'foo')))

##     def checkCreateElementNS(self):
##         el = self.document.createElementNS(TEST_NAMESPACE, 'doc')

##         assert el.hasAttributeNS(TEST_NAMESPACE, 'foo'), (
##             'Newly created Element Node should have default attribute.')
##         assert el.getAttributeNS(TEST_NAMESPACE, 'foo') == 'bar', (
##             "Wrong value of default attribute found, expected 'bar', "
##             "found %s" % el.getAttributeNS(TEST_NAMESPACE, 'foo'))
##         checkAttribute(el.getAttributeNodeNS(TEST_NAMESPACE, 'foo'),
##             'specified', 0)

##     def checkImportNodeToDefault(self):
##         newDoc = self.implementation.createDocument(None, 'baz', None)
##         newEl = newDoc.createElementNS(TEST_NAMESPACE, 'prefix:doc')
        
##         el = self.document.importNode(newEl, 0)

##         assert el.hasAttributeNS(TEST_NAMESPACE, 'foo'), (
##             'Imported Element Node should have default attribute.')
##         assert el.getAttributeNS(TEST_NAMESPACE, 'foo') == 'bar', (
##             "Wrong value of default attribute found, expected 'bar', "
##             "found %s" % repr(el.getAttributeNS(TEST_NAMESPACE, 'foo')))
##         checkAttribute(el.getAttributeNodeNS(TEST_NAMESPACE, 'foo'),
##             'specified', 0)

##     def checkChangePrefixUnspecified(self):
##         attr = self.document.documentElement.getAttributeNodeNS(TEST_NAMESPACE,
##             'foo')
##         attr.prefix = 'test'

##         # Changing the prefix of a default attribute shouldn't make a new
##         # unspecified (default) attribute node appear.
##         # TODO: It turns out this may be a hole in the spec. See PXML(60)[]:
##         # http://www.zope.org/Members/karl/ParsedXML/ParsedXMLTracker/60
##         # Waiting for consensus from DOM WG. Changing the prefix *should* make
##         # a new Attr node appear it seems. I am not convinced yet.
##         assert attr.ownerElement.attributes.length == 1, (
##             "Changing the prefix of a default attribute caused a new default "
##             "attribute node to be created.")

##         # The specified flag shouldn't change; we didn't change the value
##         checkAttribute(attr, 'specified', 0)

##     def checkChangePrefixSpecified(self):
##         self.document.documentElement.setAttributeNS(TEST_NAMESPACE, 'foo',
##             'newValue')
##         attr = self.document.documentElement.getAttributeNodeNS(TEST_NAMESPACE,
##             'foo')
##         attr.prefix = 'test'
    
##         # Changing the prefix of a specified attribute shouldn't make a new
##         # unspecified (default) attribute node appear.
##         assert attr.ownerElement.attributes.length == 2, (
##             "Changing the prefix of a specified attribute caused a new "
##             "default attribute node to be created.")

##     def checkRemoveAttributeNS(self):
##         el = self.document.documentElement
##         # Replace default with specified attr
##         el.setAttributeNS(TEST_NAMESPACE, 'foo', 'baz')

##         el.removeAttributeNS(TEST_NAMESPACE, 'foo')
##         assert el.hasAttributeNS(TEST_NAMESPACE, 'foo'), (
##             'Removing specified attribute should '
##             'bring back default attribute.')
##         assert el.getAttributeNS(TEST_NAMESPACE, 'foo') == 'bar', (
##             "Wrong value of default attribute foud, expected 'bar', "
##             "found %s" % repr(el.getAttributeNS(TEST_NAMESPACE, 'foo')))
##         checkAttribute(el.getAttributeNodeNS(TEST_NAMESPACE, 'foo'),
##             'specified', 0)

##     def checkRemoveAttributeNode(self):
##         el = self.document.documentElement
##         newAttr = self.document.createAttributeNS(TEST_NAMESPACE, 'foo')
##         newAttr.value = 'baz'
##         # Replace default with specified attr
##         el.setAttributeNodeNS(newAttr)

##         el.removeAttributeNode(newAttr)
##         assert el.hasAttributeNS(TEST_NAMESPACE, 'foo'), (
##             'Removing specified attribute should bring back default '
##             'attribute.')
##         assert el.getAttributeNS(TEST_NAMESPACE, 'foo') == 'bar', (
##             "Wrong value of default attribute found, expected 'bar', "
##             "found %s" % repr(el.getAttributeNS(TEST_NAMESPACE, 'foo')))
##         checkAttribute(el.getAttributeNodeNS(TEST_NAMESPACE, 'foo'),
##             'specified', 0)

##     def checkRemoveNamedItemNS(self):
##         el = self.document.documentElement
##         # Replace default with specified attr
##         el.setAttributeNS(TEST_NAMESPACE, 'foo', 'baz')

##         el.attributes.removeNamedItemNS(TEST_NAMESPACE, 'foo')
##         assert el.hasAttributeNS(TEST_NAMESPACE, 'foo'), \
##             'Removing specified attribute should bring back default attribute.'
##         assert el.getAttributeNS(TEST_NAMESPACE, 'foo') == 'bar', (
##             "Wrong value of default attribute found, expected 'bar', found %s"
##             % repr(el.getAttributeNS(TEST_NAMESPACE, 'foo')))
##         checkAttribute(el.getAttributeNodeNS(TEST_NAMESPACE, 'foo'),
##             'specified', 0)

##     def checkSetAttributeNS(self):
##         el = self.document.documentElement
##         el.setAttributeNS(TEST_NAMESPACE, 'foo', 'baz')
##         checkAttribute(el.getAttributeNodeNS(TEST_NAMESPACE, 'foo'),
##             'specified', 1)

##     def checkSetAttributeNodeNS(self):
##         el = self.document.documentElement
##         newAttr = self.document.createAttributeNS(TEST_NAMESPACE, 'foo')
##         newAttr.value = 'baz'
##         el.setAttributeNode(newAttr)
##         checkAttribute(el.getAttributeNodeNS(TEST_NAMESPACE, 'foo'),
##             'specified', 1)


# --- DocumentFragment

class DocumentFragmentReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        self.docfrag = self.createDocument().createDocumentFragment()
        self.node = self.docfrag

    def checkImportNode(self):
        foreignDoc = self.implementation.createDocument(None, 'foo', None)
        frag = self.docfrag

        frag.appendChild(self.document.createComment('foo'))
        frag.appendChild(self.document.createTextNode('bar'))
        
        clone = foreignDoc.importNode(frag, 0)
        deepClone = foreignDoc.importNode(frag, 1)

        assert not isSameNode(frag, clone), "Clone is same Node as original."
        assert not isSameNode(frag, deepClone), (
            "Clone is same Node as original.")

        checkAttributeSameNode(clone, 'ownerDocument', foreignDoc)
        checkAttributeSameNode(deepClone, 'ownerDocument', foreignDoc)
        checkAttribute(clone, 'parentNode', None)
        checkAttribute(deepClone, 'parentNode', None)
        checkLength(clone.childNodes, 0)
        checkLength(deepClone.childNodes, frag.childNodes.length)

        for i in range(deepClone.childNodes.length):
            checkAttribute(deepClone.childNodes.item(i), 'nodeType',
                frag.childNodes.item(i).nodeType)
            checkAttribute(deepClone.childNodes.item(i), 'data',
                frag.childNodes.item(i).data)
            checkAttributeSameNode(deepClone.childNodes.item(i),
                'ownerDocument', foreignDoc)


class DocumentFragmentWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        self.docfrag = self.createDocument().createDocumentFragment()
        self.node = self.docfrag


# --- NamedNodeMap

class NamedNodeMapWriteTestCase(TestCaseBase):

    TEST_NAMESPACE = NodeReadTestCaseBase.TEST_NAMESPACE
    TEST_PREFIX = NodeReadTestCaseBase.TEST_PREFIX
    TEST_LOCAL_NAME = NodeReadTestCaseBase.TEST_LOCAL_NAME
    TEST_QUALIFIED_NAME = NodeReadTestCaseBase.TEST_QUALIFIED_NAME

    def setUp(self):
        self.map = self.createDocument().createElement("foo")._get_attributes()
        self.attribute = self.document.createAttributeNS(self.TEST_NAMESPACE,
            self.TEST_QUALIFIED_NAME)
        self.attribute.value = "attrValue"
        self.map.setNamedItemNS(self.attribute)

    def checkGetNamedItemNS(self):
        node = self.map.getNamedItemNS(self.TEST_NAMESPACE, 
            self.TEST_LOCAL_NAME)

        assert node is not None, "getNamedItemNS didn't retrieve attribute."
        assert isSameNode(node, self.attribute), (
            "getNamedItemNS retrieved incorrect attribute.")

    def checkGetNamedItemNSWrongNamespace(self):
        node = self.map.getNamedItemNS('uri:foo', self.TEST_LOCAL_NAME)
        assert node is None, "getNamedItemNS returned an attribute."

    def checkGetNamedItemNSWrongLocalname(self):
        node = self.map.getNamedItemNS(self.TEST_NAMESPACE, 'bar')
        assert node is None, "getNamedItemNS returned an attribute."

    def checkRemoveNamedItemNS(self):
        node = self.map.removeNamedItemNS(self.TEST_NAMESPACE,
            self.TEST_LOCAL_NAME)

        assert node is not None, (
            "removeNamedItemNS didn't return an attribute.")
        assert isSameNode(node, self.attribute), (
            "removeNamedItemNS returned incorrect attribute.")
        assert self.map.getNamedItemNS(self.TEST_NAMESPACE, 
            self.TEST_LOCAL_NAME) is None, "Attribute was not removed."
        checkLength(self.map, 0)

    def checkRemoveNamedItemNSNotFound(self):
        # Exceptions
        try:
            self.map.removeNamedItemNS('uri:foo', 'bar:baz')
        except xml.dom.NotFoundErr:
            pass
        else:
            assert 0, "Removal of non-existent item succeeded."

    def checkSetNamedItemNS(self):
        newAttr = self.document.createAttributeNS(self.TEST_NAMESPACE,
            'qname:someAttr')
        newAttr.value = 'spam'
        
        retVal = self.map.setNamedItemNS(newAttr)
        assert retVal is None, "setNamedItemNS returned %s" % repr(retVal)
        checkLength(self.map, 2)
        assert isSameNode(self.map.getNamedItemNS(self.TEST_NAMESPACE,
            'someAttr'), newAttr), (
                "setNamedItemNS store seems to have failed, can't retrieve.")

    def checkSetNamedItemNSReplaceExisting(self):
        newAttr = self.document.createAttributeNS(self.TEST_NAMESPACE,
            'qname:someAttr')
        self.map.setNamedItemNS(newAttr)
        anotherAttr = self.document.createAttributeNS(self.TEST_NAMESPACE,
            'anotherQN:someAttr')
        anotherAttr.value = 'eggs'

        retVal = self.map.setNamedItemNS(newAttr)
        assert retVal is not None, "setNamedItemNS returned None"
        assert isSameNode(retVal, newAttr), (
            "setNamedItemNS didn't return replaced Node.")
        checkLength(self.map, 2)

    def checkSetNamedItemNSWrongDocument(self):
        newDoc = self.implementation.createDocument(None, 'foo', None)
        foreignAttr = newDoc.createAttributeNS(self.TEST_NAMESPACE,
            self.TEST_QUALIFIED_NAME)
        try:
            self.map.setNamedItem(foreignAttr)
        except xml.dom.WrongDocumentErr:
            pass
        else:
            assert 0, "Was allowed to add foreign Node to NamedNodeMap."

    def checkSetNamedItemNSAlreadyInUse(self):
        el = self.document.createElement('someElement')
        attr = self.document.createAttributeNS(self.TEST_NAMESPACE,
            self.TEST_QUALIFIED_NAME)
        el.setAttributeNode(attr)
        try:
           self.map.setNamedItem(attr)
        except xml.dom.InuseAttributeErr:
            pass
        else:
            assert 0, (
                "Was allowed to add attribute already in use to NamedNodeMap.")

    def checkSetNamedItemNSHierarchyRequestErr(self):
        # See DOM erratum core-4.
        element = self.document.createElementNS(TEST_NAMESPACE, 'foo:bar')
        try:
            self.map.setNamedItemNS(element)
        except xml.dom.HierarchyRequestErr:
            pass
        else:
            assert 0, (
                "Was allowed to add a Node Type not belonging in this "
                "NamedNodeMap (an Element Node to a map of attributes).")


cases = buildCases(__name__, 'Core', '2.0')



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase

ZopeTestCase.installProduct('ParsedXML')

from domapi.Base import checkAttribute

from Products.ParsedXML import ParsedXML, ExtraDOM, DOM, PrettyPrinter
from StringIO import StringIO

# the printer is a convenient way to track changes made by the parser
def printElement(element):
    output = StringIO()
    PrettyPrinter.PrintVisitor(element, output)()
    return output.getvalue()

def checkOutput(wanted, got, message="Bad output"):
    assert wanted == got, \
           ("%s.  Wanted:\n%s[[EOF]]\nGot:\n%s[[EOF]]\n" 
            % (message, wanted, got))

class ParsedXMLTestCaseBase(ZopeTestCase.ZopeTestCase):

    def shotgunParse(self, doc, namespaces = 1):
        "parse and print every node in the document"
        # first we parse a print of the document, to avoid errors from
        # comparing before and after infospace-lossyness
        docStr = StringIO(ExtraDOM.writeStream(doc).getvalue().encode('UTF-8'))
        doc = ExtraDOM.parseFile(doc, docStr, namespaces)
        from Products.ParsedXML.DOM.Traversal import NodeFilter
        nodes = []
        iterator = doc.createNodeIterator(doc, NodeFilter.SHOW_ALL, None, 0)
        # parse from the bottom up because parsing a node replaces children
        # it'd be simpler to not store a list of nodes, but this way we
        # miss possible iterator bugs
        while iterator.nextNode():
            pass
        node = iterator.previousNode()
        while node:
            if node.nodeType != DOM.Core.Node.DOCUMENT_TYPE_NODE:
                nodes.append(node)
            node = iterator.previousNode()
        # parse print of each node and compare printed doc, node.
        for node in nodes:
            docStrIn = ExtraDOM.writeStream(doc).getvalue().encode('UTF-8')
            nodeStrIn = ExtraDOM.writeStream(node).getvalue().encode('UTF-8')
            ExtraDOM.parseFile(node, StringIO(nodeStrIn), namespaces)
            nodeStrOut = ExtraDOM.writeStream(node).getvalue().encode('UTF-8')
            docStrOut = ExtraDOM.writeStream(doc).getvalue().encode('UTF-8')
            checkOutput(nodeStrIn, nodeStrOut,
                        "parsing print of node %s changed node print" % node)
            checkOutput(docStrIn, docStrOut,
                        "parsing print of node %s changed doc print" % node)

class ParseOasisXMLTestSaTestCase(ParsedXMLTestCaseBase):

    # set this to true to generate new output files
    generate = 0

    def testParse001(self):
        self._checkParse("001.xml")
    def testParse002(self):
        self._checkParse("002.xml")
    def testParse003(self):
        self._checkParse("003.xml")
    def testParse004(self):
        self._checkParse("004.xml")
    def testParse005(self):
        self._checkParse("005.xml")
    def testParse006(self):
        self._checkParse("006.xml")
    def testParse007(self):
        self._checkParse("007.xml")
    def testParse008(self):
        self._checkParse("008.xml")
    def testParse009(self):
        self._checkParse("009.xml")
    def testParse010(self):
        self._checkParse("010.xml")
    def testParse011(self):
        self._checkParse("011.xml")
    # This test fails when we use namespaces to determine valid tagnames.
    # We want to parse namespaces always.
    #def testParse012(self):
    #    self._checkParse("012.xml")
    def testParse013(self):
        self._checkParse("013.xml")
    def testParse014(self):
        self._checkParse("014.xml")
    def testParse015(self):
        self._checkParse("015.xml")
    def testParse016(self):
        self._checkParse("016.xml")
    def testParse017(self):
        self._checkParse("017.xml")
    def testParse018(self):
        self._checkParse("018.xml")
    def testParse019(self):
        self._checkParse("019.xml")
    def testParse020(self):
        self._checkParse("020.xml")
    def testParse021(self):
        self._checkParse("021.xml")
    def testParse022(self):
        self._checkParse("022.xml")
##     def testParse023(self):
##         self._checkParse("023.xml")
##     def testParse024(self):
##         self._checkParse("024.xml")
    def testParse025(self):
        self._checkParse("025.xml")
    def testParse026(self):
        self._checkParse("026.xml")
    def testParse027(self):
        self._checkParse("027.xml")
    def testParse028(self):
        self._checkParse("028.xml")
    def testParse029(self):
        self._checkParse("029.xml")
    def testParse030(self):
        self._checkParse("030.xml")
    def testParse031(self):
        self._checkParse("031.xml")
    def testParse032(self):
        self._checkParse("032.xml")
    def testParse033(self):
        self._checkParse("033.xml")
    def testParse034(self):
        self._checkParse("034.xml")
    def testParse035(self):
        self._checkParse("035.xml")
    def testParse036(self):
        self._checkParse("036.xml")
    def testParse037(self):
        self._checkParse("037.xml")
    def testParse038(self):
        self._checkParse("038.xml")
    def testParse039(self):
        self._checkParse("039.xml")
    def testParse040(self):
        self._checkParse("040.xml")
    def testParse041(self):
        self._checkParse("041.xml")
    def testParse042(self):
        self._checkParse("042.xml")
    def testParse043(self):
        self._checkParse("043.xml")
    def testParse044(self):
        self._checkParse("044.xml")
    def testParse045(self):
        self._checkParse("045.xml")
    def testParse046(self):
        self._checkParse("046.xml")
    def testParse047(self):
        self._checkParse("047.xml")
    def testParse048(self):
        self._checkParse("048.xml")
    def testParse049(self):
        self._checkParse("049.xml")
    def testParse050(self):
        self._checkParse("050.xml")     
##     def testParse051(self):
##        self._checkParse("051.xml")
    def testParse052(self):
        self._checkParse("052.xml")
##     def testParse053(self):
##         self._checkParse("053.xml")
    def testParse054(self):
        self._checkParse("054.xml")
    def testParse055(self):
        self._checkParse("055.xml")
    def testParse056(self):
        self._checkParse("056.xml")
    def testParse057(self):
        self._checkParse("057.xml")
    def testParse058(self):
        self._checkParse("058.xml")
    def testParse059(self):
        self._checkParse("059.xml")
    def testParse060(self):
        self._checkParse("060.xml")
    def testParse061(self):
        self._checkParse("061.xml")
    def testParse062(self):
        self._checkParse("062.xml")
##     def testParse063(self):
##         self._checkParse("063.xml")
    def testParse064(self):
        self._checkParse("064.xml")
##     def testParse065(self):
##         self._checkParse("065.xml")
##     def testParse066(self):
##         self._checkParse("066.xml")
    def testParse067(self):
        self._checkParse("067.xml")
##     def testParse068(self):
##         self._checkParse("068.xml")
##     def testParse069(self):
##          self._checkParse("069.xml")
    def testParse070(self):
        self._checkParse("070.xml")
    def testParse071(self):
        self._checkParse("071.xml")
    def testParse072(self):
        self._checkParse("072.xml")
    def testParse073(self):
        self._checkParse("073.xml")
    def testParse074(self):
        self._checkParse("074.xml")
    def testParse075(self):
        self._checkParse("075.xml")
##     def testParse076(self):
##         self._checkParse("076.xml")
    def testParse077(self):
        self._checkParse("077.xml")
    def testParse078(self):
        self._checkParse("078.xml")
    def testParse079(self):
        self._checkParse("079.xml")
    def testParse080(self):
        self._checkParse("080.xml")
    def testParse081(self):
        self._checkParse("081.xml")
    def testParse082(self):
        self._checkParse("082.xml")
    def testParse083(self):
        self._checkParse("083.xml")
    def testParse084(self):
        self._checkParse("084.xml")
##     def testParse085(self):
##         self._checkParse("085.xml")
##     def testParse086(self):
##         self._checkParse("086.xml")
##     def testParse087(self):
##         self._checkParse("087.xml")
##     def testParse088(self):
##         self._checkParse("088.xml")
##     def testParse089(self):
##         self._checkParse("089.xml")
##     def testParse090(self):
##         self._checkParse("090.xml")
##     def testParse091(self):
##         self._checkParse("091.xml")
    def testParse092(self):
        self._checkParse("092.xml")
    def testParse093(self):
        self._checkParse("093.xml")
    def testParse094(self):
        self._checkParse("094.xml")
    def testParse095(self):
        self._checkParse("095.xml")
    def testParse096(self):
        self._checkParse("096.xml")
    def testParse097(self):
        self._checkParse("097.xml")
    def testParse098(self):
        self._checkParse("098.xml")
    def testParse099(self):
        self._checkParse("099.xml")
##     def testParse100(self):
##         self._checkParse("100.xml")
##     def testParse101(self):
##         self._checkParse("101.xml")
    def testParse102(self):
        self._checkParse("102.xml")
    def testParse103(self):
        self._checkParse("103.xml")
    def testParse104(self):
        self._checkParse("104.xml")
    def testParse105(self):
        self._checkParse("105.xml")
    def testParse106(self):
        self._checkParse("106.xml")
    def testParse107(self):
        self._checkParse("107.xml")
##     def testParse108(self):
##         self._checkParse("108.xml")
    def testParse109(self):
        self._checkParse("109.xml")
##     def testParse110(self):
##         self._checkParse("110.xml")
    def testParse111(self):
        self._checkParse("111.xml")
    def testParse112(self):
        self._checkParse("112.xml")
    def testParse113(self):
        self._checkParse("113.xml")
##     def testParse114(self):
##         self._checkParse("114.xml")
##     def testParse115(self):
##         self._checkParse("115.xml")
##     def testParse116(self):
##         self._checkParse("116.xml")
##     def testParse117(self):
##         self._checkParse("117.xml")
##     def testParse118(self):
##         self._checkParse("118.xml")

    def testParse119(self):
        self._checkParse("119.xml")

    def _checkParse(self, iterFileName):
        testDir = os.path.join(sys.modules['Products.ParsedXML'].__path__[0],
            'tests')
        saDir = os.path.join(testDir, 'xml', 'conf', 'xmltest', 'sa')
        outDir = os.path.join(saDir, 'ParsedXMLTestOut')

        # Read the test input
        inFilename = os.path.join(saDir, iterFileName)
        inFile = open(inFilename).read()
        doc = ParsedXML.ParsedXML('foo', inFile)

        # Read the output to test against
        outFilename = os.path.join(outDir, iterFileName)
        outFile = open(outFilename).read()
        # FIXME: if the next line is enabled, the tests will succeed
        # with python 2.1. Unfortunately the whole testsuite will
        # segfault at about test 23 (it'll vary)
        outFile = unicode(outFile, 'utf-8')
        
        # Print the DOM, and compare against the expected output.
        output = printElement(doc)
        
        # All line separators are supposed to normalize to Unix-style
        # (XML 1.0, section 2.11).
##         outFile = string.replace(outFile, "\r\n", "\n")
##         outFile = string.replace(outFile, "\r", "\n")

        if self.generate:
            # re-generate the expected output file, but only if it changed
            if outFile != output:
                fp = open(outFilename, "w")
                fp.write(output)
                fp.close()
        else:
            checkOutput(outFile, output)
            self.shotgunParse(doc.getDOMObj())

class ParseTestCase(ParsedXMLTestCaseBase):

    def testParseException(self):
        "assert exception & exception args are correct"
        from xml.parsers import expat
        text = "<doc>\nfoobar<</doc>" # parse error line 2 column 7
        try:
            ParsedXML.ParsedXML('foo', text)
        except expat.error, e:
            assert e.lineno == 2, (
                "parse exception give wrong line number.  Wanted %s got %s"
                % (2, e.lineno))
            assert e.offset == 7, (
                "parse exception give wrong column number.  Wanted %s got %s"
                % (7, e.offset))
        else:
            assert 0, "parse of malformed XML doesn't raise properly"

    def testSubnodeParseException(self):
        "assert exception & exception args are correct"        
        from xml.parsers import expat
        docText = "<doc><child/></doc>"
        subText = "<child>\nfoobar<</child>" # parse error line 2 column 7
        doc = ParsedXML.ParsedXML('foo', docText)
        try:
            doc.documentElement.firstChild.parseXML(StringIO(subText))
        except expat.error, e:
            assert e.lineno == 2, (
                "parse exception gives wrong line number.  Wanted %s, got %s"
                % (2, e.lineno))
            assert e.offset == 7, (
                "parse exception gives wrong offset.  Wanted %s, got %s"
                % (7, e.offset))
        else:
            assert 0, "parse of malformed XML doesn't raise properly"

class Lvl2ParseTestCase(ParsedXMLTestCaseBase):

    def testNamespaceAttrOfDocumentElement(self):
        """we should be able to parse an element that uses a namespace
        declared on the element itself"""
        inStr = '<?xml version="1.0" ?>\n' \
                '<foo:bar xmlns:foo="uri:test_namespace"/>\n'
        doc = ParsedXML.ParsedXML('foo', inStr)
        self.shotgunParse(doc.getDOMObj())        

    def testNamespaceAttrOfElement(self):
        """we should be able to parse an element that uses a namespace
        declared on the element itself"""
        inStr = '<?xml version="1.0" ?>\n' \
                '<doc><foo:bar xmlns:foo="uri:test_namespace"/></doc>\n'
        doc = ParsedXML.ParsedXML('foo', inStr)
        self.shotgunParse(doc.getDOMObj())        

    def testSubnodeAncestorNamespace(self):
        """we should be able to parse a subtree that uses a namespace
        declared on an ancestor that we don't parse"""
        inStr = ('<?xml version="1.0" ?>\n'
                 '<doc><spam xmlns:spamNs="uri:test_namespace">'
                 '<eggs><spamNs:ham spamNs:foo="bar"/></eggs>'
                 '</spam></doc>\n')
        doc = ParsedXML.ParsedXML('foo', inStr)
        self.shotgunParse(doc.getDOMObj())

    def testSubnodeParseXMLNamepsaceDecl(self):
        """Check that we can parse at a subnode with an xml ns decl attr, and
        that parsing a subnode's output doesn't change the document.
        The external entity parser that the fragment builder uses likes
        to add this namespace, so we want to make sure we don't break
        anything that uses it."""
        inStr = '<spam xmlns:xml="http://www.w3.org/XML/1998/namespace"/>'
        doc = ParsedXML.ParsedXML('foo', inStr)
        self.shotgunParse(doc.getDOMObj())

    def testSubnodeParseXMLNamepsace(self):
        """Check that we can parse at a subnode with an xml ns attr, and
        that parsing a subnode's output doesn't change the document.
        The external entity parser that the fragment builder uses likes
        to add this namespace, so we want to make sure we don't break
        anything that uses it."""
        # we should use a real XML name here, it's not an error if the
        # parser notices that we're abusign the xml namespace
        inStr = '<spam xml:spam="spam"/>'
        doc = ParsedXML.ParsedXML('foo', inStr)
        self.shotgunParse(doc.getDOMObj())

    def testXMLNSPrefixParse(self):
        "check that xmlns prefix attrs are parsed properly"
        inStr = '<doc xmlns:spamNS="uri:test_namespace"/>'
        doc = ParsedXML.ParsedXML('foo', inStr)
        attr = doc.documentElement.attributes.item(0)
        checkAttribute(attr, 'prefix', 'xmlns')
        checkAttribute(attr, 'localName', 'spamNS')
        checkAttribute(attr, 'namespaceURI', 'http://www.w3.org/2000/xmlns/')
        checkAttribute(attr, 'value', 'uri:test_namespace')        

    def testXMLNSParse(self):
        "check that xmlns attrs are parsed properly"
        inStr = '<doc xmlns="uri:test_namespace"/>'
        doc = ParsedXML.ParsedXML('foo', inStr)
        attr = doc.documentElement.attributes.item(0)
        checkAttribute(attr, 'prefix', 'xmlns')
        checkAttribute(attr, 'localName', None)
        checkAttribute(attr, 'namespaceURI', 'http://www.w3.org/2000/xmlns/')
        checkAttribute(attr, 'value', 'uri:test_namespace')        

    def testNoNSParse(self):
        """check that attrs parsed with no NS but with NS aware parse
        can be retrieved with NS of None"""
        inStr = '<doc version="1.0"/>'
        doc = ParsedXML.ParsedXML('foo', inStr)
        assert doc.documentElement.getAttributeNS(None, 'version'), (
            "NS-free attribute not gotten by NS-free getAttributeNS")

if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(ParseOasisXMLTestSaTestCase))
        suite.addTest(unittest.makeSuite(ParseTestCase))
        suite.addTest(unittest.makeSuite(Lvl2ParseTestCase))
        return suite





##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

"""A suite of unit tests for the Python DOM API.

This suite will test a DOM API for compliance with the DOM API, level 2. It
assumes that the DOM tested supports at least both the Core and XML features. It
requires PyUnit; see http://pyunit.sourceforge.net/.

Example for the python minidom (which is an incomplete implementation):

from xml.dom.minidom import DOMImplementation, parseString
from domapi import DOMImplementationTestSuite

def MiniDomParseString(self, xml):
    return parseString(xml)

def test_suite():
    '''Return a test suite for the Zope testing framework.'''
    return DOMImplementationTestSuite(DOMImplementation(), MiniDomParseString)

if __name__ == '__main__':
    import unittest
    unittest.TextTestRunner().run(test_suite())

"""

import unittest
import CoreLvl1, CoreLvl2, CoreLvl3, XMLLvl1, XMLLvl2, TraversalLvl2, Load3

cases = (
    CoreLvl1.cases +
    CoreLvl2.cases +
    CoreLvl3.cases +
    XMLLvl1.cases +
    XMLLvl2.cases +
    TraversalLvl2.cases +
    Load3.cases
)

def DOMImplementationTestSuite(implementation, parseMethod, verbose=0):
    """ Create a testsuite for DOM lvl 2 compliance, given a DOM Implementation.

    To test a DOM implementation, hand in a DOMImplementation object, and a
    method that will take a string holding an XML document and returns a
    Document Node created from the XML.

    Then run the returned unittest testsuite.

    Note that the signature of the parse method is (self, xmlString) and should
    parse the xml string with namespaces turned on. It will be used to create
    Nodes which normally cannot be created using the DOM API, like Notations and
    Entities.

    """

    # First test for minimal feature support
    assert (implementation.hasFeature('Core', '2.0') and
        implementation.hasFeature('XML', '2.0')), (
        "This DOMImplementation doesn't feature the level 2 Core and XML API.")

    suite = unittest.TestSuite()
    # The minimal set of features that should return 1 on hasFeature. 
    # ('Core', '1.0') was never defined for DOM level 1, it was implicit. Most
    # DOM level 2 implementations return true, but this is a courtesy.
    supportedFeatures = {
        ('Core', None): 1,
        ('Core', '2.0'): 1,
        ('XML', None): 1,
        ('XML', '1.0'): 1,
        ('XML', '2.0'): 1,
    }

    for case, feature, version in cases:
        if implementation.hasFeature(feature, version):
            case.implementation = implementation
            case.parse = parseMethod
            suite.addTest(unittest.makeSuite(case, 'check'))

            supportedFeatures[(feature, version)] = 1
            supportedFeatures[(feature, None)] = 1
        else:
            if verbose:
                 print ("Test %s skipped: DOM feature not supported.\n" %
                    case.__name__)

    # Record supported features.
    CoreLvl1.DOMImplementationReadTestCase.supportedFeatures = (
        supportedFeatures.keys())
    CoreLvl2.NodeReadTestCaseBase.supportedFeatures = (
        supportedFeatures.keys())

    return suite



##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Serializers for Shane's AdaptableStorage

    The properties must be deserialized prior the actual xml.

$Id: AdaptableStorage.py,v 1.2 2004/04/27 13:06:18 faassen Exp $
"""
from types import StringType
from cStringIO import StringIO

from Products.AdaptableStorage.mapper.public import IAspectSerializer,\
    FieldSchema, RowSequenceSchema
from ParsedXML import ParsedXML, createDOMDocument

class ParsedXMLSerializer:
    """Serializer for ParsedXML Documents.

        Only the raw XML is stored in the filesystem.
    """
    
    __implements__ = IAspectSerializer

    schema = FieldSchema('data', 'string')
    attributes = (
           )
        
    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        return isinstance(object, ParsedXML)
    
    def serialize(self, object, event):
        encoding = 'utf-8'
        data = object.writeStream(encoding=encoding).getvalue().encode(encoding)
        event.ignoreAttributes([ '_node', '_persistentDoc', '_lenCache',
            '_lenCorrect', '_container'])
        return data

    def deserialize(self, object, event, state):
        assert isinstance(state, StringType)
        object._persistentDoc = object
        object._lenCache = 0
        object._lenCorrect = 0
        object.initFromDOMDocument(createDOMDocument(state, 
            not object.noNamespaces))


class ParsedXMLPropertiesSerializer:
    """Serialize parsed xml's properties"""
    
    __implements__ = IAspectSerializer
    
    schema = RowSequenceSchema()
    schema.addField('id', 'string', 1)
    schema.addField('type', 'string')
    schema.addField('data', 'string')

    attributes = {
        'title': str,
        'contentType': str,
        'noNamespaces': int,
     }

    def getSchema(self):
        return self.schema

    def canSerialize(self, object):
        return isinstance(object, ParsedXML)

    def serialize(self, object, event):
        res = []
        for attribute, factory in self.attributes.items():
            if not hasattr(object, attribute):
                continue
            value = getattr(object, attribute)
            t = factory.__name__
            if value is None:
                if factory in (int, long):
                    value = 0
                else: 
                    value = ''
            value = str(value)
            event.ignoreAttribute(attribute)
            res.append((attribute, t, value))
        return res 

    def deserialize(self, object, event, state):
        for attribute, t, value in state:
            factory = self.attributes.get(attribute)
            if factory is None:
                continue
            value = factory(value)
            setattr(object, attribute, value)
            assert getattr(object, attribute) == value




##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import unittest
# needed to import NodePath module
import sys
sys.path.insert(0, '../..')

from NodePath import registry

def DOMParseString(xml):
    # FIXME: change this if using another DOM
    from StringIO import StringIO
    from Products.ParsedXML.DOM import ExpatBuilder
    file = StringIO(xml)
    return ExpatBuilder.parse(file)

class NodePathTestCase(unittest.TestCase):
    def setUp(self):
        doc = DOMParseString('''
        <doc>
        <chapter title="one">
        <p>This is a very trival <i>XML</i> document.</p>
        <p>We will test whether our node path facility works with it.</p>
        </chapter>
        <chapter title="two">
        <p>Here is a second chapter, which contains a list.</p>
        <list>
        <element>Foo</element>
        <element>Bar</element>
        <element>Baz</element>
        </list>
        </chapter>
        </doc>
        ''')
        self._doc = doc
        
    def _shotgun_check(self, top_node, node, scheme_name):
        path = registry.create_path(top_node, node, scheme_name)
        found = registry.resolve_path(top_node, path)
        assert found == node, ("Found %s with '%s', wanted %s" %
                               (found, path, node))
        cycled_path = registry.create_path(top_node, node, scheme_name)
        assert path == cycled_path, ("Cycled path %s found, wanted %s" %
                                     (cycled_path, path))
        for child in node.childNodes:
            self._shotgun_check(top_node, child, scheme_name)

    def _shotgun_check_robust(self, top_node, node):
        path = registry.create_path(top_node, node, 'robust')
        found = registry.resolve_path(top_node, path)
        assert found == node, ("Found %s with '%s', wanted %s" %
                               (found, path, node))
        # can't cycle path as words are selected randomly
        for child in node.childNodes:
            self._shotgun_check_robust(top_node, child)
  
    def checkChildPath(self):
        self._shotgun_check(self._doc, self._doc.documentElement, 'child')
        
    def checkElementIdPath(self):
        self._shotgun_check(self._doc, self._doc.documentElement, 'element_id')
        
    def checkRobustPath(self):
        self._shotgun_check_robust(self._doc, self._doc.documentElement)

    def checkEmptyPath(self):
        self.assertEquals('',
                          registry.create_path(self._doc, self._doc, 'child'))
        self.assertEquals('',
                          registry.create_path(self._doc, self._doc, 'element_id'))     
        #self.assertEquals('',
        #                  registry.create_path(self._doc, self._doc, 'robust'))
        found = registry.resolve_path(self._doc, '')
        self.assertEquals(self._doc, found)
        
def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(NodePathTestCase, "check"))
    return suite
    
def main():
    unittest.TextTestRunner().run(test_suite())

if __name__ == "__main__":
    main()
    



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from NodePath import BaseNodePathScheme, NodePathError
import random, string, urllib
import xml.dom

words_amount = 5

class RobustPathScheme(BaseNodePathScheme):
    def __init__(self):
        BaseNodePathScheme.__init__(self, 'robust')

    def resolve_steps(self, top_node, steps):
        steps, type_step, context_step = steps[:-2], steps[-2], steps[-1]
        rsteps = []
        # create element steps
        for step in steps:
            nodeName, nodeIndex = string.split(step, "*")
            rsteps.append(GenericStep(xml.dom.Node.ELEMENT_NODE,
                                      nodeName, int(nodeIndex)))
        # now create type and context end steps
        context_parts = string.split(context_step, "*")
        node_type, nodeIndex = string.split(type_step, "*")
        nodeIndex = int(nodeIndex)
        if node_type == 'element':
            type_rstep = GenericStep(xml.dom.Node.ELEMENT_NODE,
                                context_parts[0], nodeIndex)
            context_rstep = ElementEndStep(int(context_parts[1]),
                                           int(context_parts[2]))
        elif node_type == 'text':
            type_rstep = GenericStep(xml.dom.Node.TEXT_NODE,
                                     "#text", nodeIndex)
            words = []
            for i in range(0, len(context_parts), 2):
                words.append((urllib.unquote(context_parts[i]),
                             int(context_parts[i + 1])))
            context_rstep = TextEndStep(words)
        elif node_type == 'empty':
            type_rstep = GenericStep(xml.dom.Node.TEXT_NODE,
                                     "#text", nodeIndex)
            context_rstep = EmptyTextEndStep()
        elif node_type == 'other':
            raise NodePathError, "Cannot handle other path elements yet."
        else:
            raise NodePathError, "Unknown step type in path: %s" % node_type

        rsteps.append(type_rstep)
        rsteps.append(context_rstep)

        # resolve the steps, starting with step 0
        context = Context(0, rsteps)
        node, level = rsteps[0].resolve(context, top_node, 0)
        return node
        
    def create_steps(self, top_node, node):
        # first create last two steps
        parent = node.parentNode
        steps = self.create_end_steps(parent, node)
        node = parent
        # now create element steps
        while node is not top_node:
            parent = node.parentNode
            if parent is None:
                break
            steps.append(self.create_element_step(parent, node))
            node = parent
        steps.reverse()
        return steps
    
    def create_end_steps(self, parent, node):
        nodeType = node.nodeType
        if nodeType == node.ELEMENT_NODE:
            return ['%s*%s*%s' % (node.nodeName,
                                  len(node.attributes),
                                  len(node.childNodes)),
                    'element*%s' % parent.childNodes.index(node) ]
        elif nodeType == node.TEXT_NODE:
            if string.strip(node.data) == '':
                return ['empty',
                        'empty*%s' % parent.childNodes.index(node)]
            else:
                return [string.join(get_words_sample(node.data), "*"),
                        'text*%s' % parent.childNodes.index(node)]
        else:
            # FIXME: simplistic way to deal with other nodes
            return ['%s' % node.nodeType,
                    'other*%s' % parent.childNodes.index(node)]
        
    def create_element_step(self, parent, node):
        return "%s*%s" % (node.nodeName, parent.childNodes.index(node))
    
class Context:
    threshold = 30
    success_threshold = 10
    offset_unreliability = 3
    tree_skip_unreliability = 3
    step_skip_unreliability = 3
    word_max_distance = 7
    word_offset_unreliability = 2
    
    def __init__(self, i, steps):
        self._i = i
        self._steps = steps
            
    def resolve_next(self, node, level):
        steps = self._steps
        i = self._i + 1
        if i < len(steps):
            # prepare new context for next step
            context = Context(i, steps)
            # resolve the next step
            return steps[i].resolve(context, node, level)
        else:
            return node, level

class GenericStep:
    def __init__(self, nodeType, nodeName, index):
        self._nodeType = nodeType
        self._nodeName = nodeName
        self._index = index
    
    def resolve(self, context, node, level):
        nodeType = self._nodeType
        nodeName = self._nodeName
        i = self._index
        childNodes = node.childNodes
        l = len(childNodes)

        results = []
        # try if the indicate node is the one, if so, return it
        if i < l:
            current = childNodes[i]
            if (current.nodeType == nodeType and
                current.nodeName == nodeName):
                 current_node, current_unreliability = context.resolve_next(
                     current, level)
                 if current_unreliability < context.success_threshold:
                     return current_node, current_unreliability
                 results.append((current_node, current_unreliability))
            # we need to go forward and backward from position i
            forward_i = i + 1
            backward_i = i - 1
            can_go_forward = can_go_backward = 1
        else:
            # i is beyond length, so we need to go backward from end
            backward_i = l - 1
            forward_i = l
            can_go_forward = 0
            can_go_backward = 1

        # try nodes forward and backward of this one
        unreliability = level + context.offset_unreliability
        while ((can_go_forward or can_go_backward) and
               unreliability < context.threshold):
            if forward_i < l:
                current = childNodes[forward_i]
                if (current.nodeType == nodeType and
                    current.nodeName == nodeName):
                    current_node, current_unreliability = context.resolve_next(
                        current, unreliability)
                    if current_unreliability < context.success_threshold:
                        return current_node, current_unreliability
                    results.append((current_node, current_unreliability))
                forward_i = forward_i + 1
            else:
                can_go_forward = 0
                
            if backward_i >= 0:
                current = childNodes[backward_i]
                if (current.nodeType == nodeType and
                    current.nodeName == nodeName):
                    current_node, current_unreliability = context.resolve_next(
                        current, unreliability)
                    if current_unreliability < context.success_threshold:
                        return current_node, current_unreliability
                    results.append((current_node, current_unreliability))   
                backward_i = backward_i - 1
            else:
                can_go_backward = 0

            unreliability = unreliability + context.offset_unreliability

        # try skipping level in the tree
        unreliability = level + context.tree_skip_unreliability
        if unreliability < context.threshold:
            for current in childNodes:
                # use same step but with next nodes
                current_node, current_unreliability = self.resolve(
                    context, current, unreliability)
                if current_unreliability < context.success_threshold:
                    return current_node, current_unreliability
                results.append((current_node, current_unreliability))  

        # try skipping this step
        # use same node but with next step
        unreliability = level + context.step_skip_unreliability
        if unreliability < context.threshold:
            current_node, current_unreliability = context.resolve_next(
                node, unreliability)
            if current_unreliability < context.success_threshold:
                return current_node, current_unreliability
            results.append((current_node, current_unreliability))
            
        # no immediate success, so try the best branch
        best_unreliability = context.threshold
        best_node = None
        for found_node, found_unreliability in results:
            if found_unreliability < best_unreliability:
                best_unreliability = found_unreliability
                best_node = found_node
        return best_node, best_unreliability
    
class ElementEndStep:
    """Not very robust, but representable in a url pretty easily.
    """
    def __init__(self, attributeAmount, childAmount):
        self._attributeAmount = attributeAmount
        self._childAmount = childAmount

    def resolve(self, context, node, level):
        if node.nodeType != node.ELEMENT_NODE:
            return None, context.threshold
        if (len(node.attributes) != self._attributeAmount or 
            len(node.childNodes) != self._childAmount):
            return None, context.threshold
        return node, level

class EmptyTextEndStep:
    
    def resolve(self, context, node, level):
        if node.nodeType != node.TEXT_NODE:
            return None, context.threshold
        if string.strip(node.data) == "":
            return node, level
        return None, context.threshold
    
class TextEndStep:
    def __init__(self, words):
        self._words = words
   
    def resolve(self, context, node, level):
        if node.nodeType != node.TEXT_NODE:
            return None, context.threshold
        words = string.split(node.data)
        if len(words) == 0:
            return None, context.threshold

        unreliability = level
        for word, nr in self._words:
            unreliability = unreliability + (
                find_word_distance(words, word, nr,
                                   context.word_max_distance) *
                context.word_offset_unreliability )
            if unreliability >= context.threshold:
                return None, context.threshold
        return node, unreliability

def find_word_distance(words, word, i, max_distance):
    """Find distance of word in words from expected location i.
    If word could not be found or is further than max_distance, return -1.
    """
    try:
        if words[i] == word:
            return 0
    except IndexError:
        l = len(words)
        distance = i - l + 1
        forward_i = l
        backward_i = l - 1
        can_go_forward = 0
        can_go_backward = 1
    else:
        l = len(words)
        distance = 1
        forward_i = i + 1
        backward_i = i - 1
        can_go_forward = can_go_backward = 1
    while (can_go_forward or can_go_backward) and (distance < max_distance):
        if forward_i < l:
            if words[forward_i] == word:
                return distance
            forward_i = forward_i + 1
        else:
            can_go_forward = 0
        if backward_i >= 0:
            if words[backward_i] == word:
                return distance
            backward_i = backward_i - 1
        else:
            can_go_backward = 0

        distance = distance + 1

    return max_distance

def get_words_sample(text,
                     words_amount=words_amount,
                     quote=urllib.quote):
    words = string.split(text)
    result = []

    if len(words) <= words_amount:
        # get all words if there are only a few
        for i in range(len(words)):
            # don't add any words with separator in it, so as not to confuse
            if "*" not in words[i] and "," not in words[i]:
                qword = quote(words[i])
                result.append(qword + "*" + str(i))
    else:
        # otherwise, take a random sample of words from the text
        nrs = []
        l = len(words) - 1
        for i in range(words_amount):
            while 1:
                r = random.randint(0, l)
                # if we selected new word from sample, done
                # NOTE: would introduce subtle bug if we checked for
                # * or , here
                if r not in nrs:
                    break
            nrs.append(r)
        nrs.sort()
        for nr in nrs:
            # don't want any word with separator in it
            if "*" not in words[nr] and "," not in words[nr]:
                qword = quote(words[nr])
                result.append(qword + "*" + str(nr))
    return result
